"""
Copyright (C) 2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

"""
from .problem.problem import Problem
from .problem.benchmark import benchmark
from .optimizer.optimizer import Optimizer
from .version import __version__
# Try to import matplotlib packages. Issue a warning if any error occurs.
import warnings
try:
    import matplotlib.pyplot as plt
    from matplotlib.ticker import MaxNLocator
    from matplotlib import cm
    from mpl_toolkits.mplot3d import Axes3D
except Exception as e:
    warnings.warn('Unable to import modules from matplotlib: %s\nAs a result, the visualization capabilities in the prosrs package may be impaired. For the workarounds and the implications of this, please see the installation note in the README file at `https://github.com/compdyn/ProSRS`.' % str(e))


"""
Copyright (C) 2016-2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

Define an optimization problem.
"""
import sys
import numpy as np
from ..utility.functions import eval_func
try:
    import matplotlib.pyplot as plt
    from matplotlib import cm
    from mpl_toolkits.mplot3d import Axes3D # needed for ``fig.add_subplot(111, projection='3d')``
except:
    pass


class Problem:
    """
    An optimization problem class that works with ProSRS algorithm.
    """
    def __init__(self, func, domain, x_var=None, y_var='y', name='demo', true_func=None, 
                 min_loc=None, min_true_func=None, noise_type=None, sd=None):
        """
        Constructor. 
        
        Args:
            
            func (callable): Function to be optimized over, for which the input is 1d array,
                the output is a scalar. For example, ``func = lambda x: x[0]**2+x[1]**2``.
            
            domain (list of tuples): Optimization domain.
                For example, `domain` = [(0, 1), (0, 2)] means optimizing the
                2D function `func` of argument ``x`` on the domain: 
                ``0 <= x[0] <= 1`` and ``0 <= x[1] <= 2``.
                
            x_var (list of str or None, optional): Argument names of function `func`.
                If None, then `x_var` defaults to ['x1', 'x2', ...]
                such that length of `x_var` = length of `domain`.
            
            y_var (str, optional): Response variable name of function `func`.
            
            name (str, optional): Name of this optimization problem.
            
            true_func (callable or None, optional): Underlying true function of
                `func` (i.e., expectation of function `func`). If None,
                then the underlying true function is unknown.
            
            min_loc (2d array or None, optional): Global minimum locations of function 
                `true_func` on the `domain`. Each row defines one minimum point. 
                If None, then the minimum locations are unknown.
            
            min_true_func (float or None, optional): Global minimum of function 
                `true_func` on the `domain`. If None, then global minimum is unknown.
            
            noise_type (str or None, optional): Type of random noises (e.g., 'Gaussian').
                If None, then the noise type is unknown.
            
            sd (float or None, optional): Standard deviation of random noise in `func`.
                If None, then the standard deviation is unknown.
        
        Raises:
            
            ValueError: If dimension of `domain` and dimension of `x_var` do not match.
        
        """
        self.f = func
        self.domain = domain
        self.dim = len(domain) # dimension of optimization problem
        self.x_var = ['x%d' % i for i in range(1, self.dim+1)] if x_var is None else x_var
        self.y_var = y_var
        self.name = name
        self.F = true_func
        self.min_loc = min_loc
        self.min_true_func = min_true_func
        self.noise_type = noise_type
        self.sd = sd
        self.domain_lb, self.domain_ub = zip(*self.domain) # domain lower bounds and domain upper bounds
        self.domain_lb, self.domain_ub = np.array(self.domain_lb), np.array(self.domain_ub) # convert to 1d array
        # sanity check
        if self.dim != len(self.x_var):
            raise ValueError('Inconsistent dimension for x_var and domain.')
        if self.min_loc is not None:
            assert(self.min_loc.shape[1] == self.dim), 'Wrong shape for min_loc.'
        if self.sd is not None:
            assert(self.sd >= 0)
        assert(np.all(self.domain_lb < self.domain_ub))
    
    
    def __str__(self):
        """
        Display optimization problem info.
        
        Use ``print()`` method to call this function.
        """
        line = 'Optimization problem:\n'
        line += '- Name: %s\n' % self.name
        line += '- Dimension: %d\n' % self.dim
        line += '- Domain: %s\n' % ('{'+', '.join(["'%s': %s" % (x, str(tuple(v))) for x, v in zip(self.x_var, self.domain)])+'}')
        line += "- Response variable: '%s'\n" % str(self.y_var)
        if self.sd is not None and self.noise_type is not None:
            line += '- Random noise: %s with standard deviation of %g\n' % (self.noise_type, self.sd)
        if self.min_true_func is not None:
            line += '- Global minimum: %g\n' % self.min_true_func
        else:
            line += '- Global minimum: unknown\n'
        if self.min_loc is not None:
            line += '- Global minimum locations: %s\n' % ', '.join(["%s = %s" % (str(tuple(self.x_var)), str(tuple(v)))\
                                                                   for v in self.min_loc])
        else:
            line += '- Global minimum locations: unknown\n'
        line = line.rstrip() # remove \n at the end
        return line
    
        
    def visualize(self, true_func=False, n_samples=None, plot='contour', 
                  contour_levels=100, min_marker_size=10, n_proc=1, fig_path=None):
        """
        Visualize optimization function (only if dimension of function is <= 2).
        
        Args:
            
            true_func (bool, optional): Whether we plot the underlying expectation 
                function `true_func`. If False, then plot (noisy) function `func`. 
            
            n_samples (int or None, optional): Number of samples for generating
                plots. If None, `n_samples` assumes default values: 20 for 1D 
                problems; 100 for 2D problems.
            
            plot (str, optional): Plot style for 2D problems. Must be one 
                of ['contour', 'surface'].
                
            contour_levels (int, optional): Number of contour lines in the contour plot.
                Useful only when `plot` = 'contour'. For more explanation, see
                `Matplotlib Document <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.contourf.html>`.
            
            min_marker_size (int, optional): Marker size for global minimum locations.
            
            n_proc (int, optional): Number of parallel processes for parallel function 
                evaluations. If = 1, then optimization function is evaluated in serial.
            
            fig_path (str, optional): Path of figure (i.e., where to save the plot). 
                If None, then we do not save visualization.
        
        Raises:
            
            TypeError: If function to be plotted is not callable.
            ValueError: If dimension is 2 and `plot` is not one of ['contour', 'surface'] or 
                        if dimension of function is greater than 2.           
        """
        if self.dim > 2:
            raise ValueError('Visualization for problems of dimension greater than 2 is not implemented.')
        else:
            try:
                plot_f = self.F if true_func else self.f # function to be plotted
                # sanity check
                if not callable(plot_f):
                    raise TypeError('Optimization function to be plotted is not callable.')
                
                fig = plt.figure()
                
                if self.dim == 1:
                    
                    n_samples = 20 if n_samples is None else n_samples
                    X = np.linspace(self.domain[0][0], self.domain[0][1], n_samples)
                    Y = eval_func(plot_f, X.reshape((-1, 1)), n_proc=n_proc)
                    plt.plot(X, Y, 'b-')
                    if self.min_loc is not None:
                        Ymin = [plot_f(x) for x in self.min_loc]
                        plt.plot(self.min_loc, Ymin, 'rx', markersize=min_marker_size)
                    
                else: # i.e., self.dim == 2
                    
                    n_samp_per_dim = 10 if n_samples is None else int(n_samples**0.5) # number of samples per dimension
                    n_samples = n_samp_per_dim**2
                    # generate and evaluate samples
                    x1 = np.linspace(self.domain[0][0], self.domain[0][1], n_samp_per_dim)
                    x2 = np.linspace(self.domain[1][0], self.domain[1][1], n_samp_per_dim)
                    x1, x2 = np.meshgrid(x1, x2)
                    X = np.hstack((x1.reshape((n_samples, 1)), x2.reshape((n_samples, 1))))
                    Y = eval_func(plot_f, X, n_proc=n_proc).reshape((n_samp_per_dim, n_samp_per_dim))
                    
                    if plot == 'contour':
                            
                        plt.contourf(x1, x2, Y, contour_levels)
                        if self.min_loc is not None:    
                            plt.plot(self.min_loc[:,0], self.min_loc[:,1], 'w.', markersize=min_marker_size)
                        plt.colorbar()
                        plt.xlabel(self.x_var[0])
                        plt.ylabel(self.x_var[1])
                        plt.title('Contour plot of optimization function (problem: %s)' % self.name)
                    
                    elif plot == 'surface':
    
                        ax = fig.add_subplot(111, projection='3d')
                        ax.plot_surface(x1, x2, Y, rstride=1, cstride=1, cmap=cm.rainbow,
                                        linewidth=0, antialiased=False)
                        ax.axis('off')
                        ax.set_xlabel(self.x_var[0])
                        ax.set_ylabel(self.x_var[1])
                        ax.set_title('Surface plot of optimization function (problem: %s)' % self.name)
                        
                    else:
                        raise ValueError("Invalid plot value. Must be one of ['contour', 'surface']")
                    
                plt.show()
                if fig_path is not None:
                    fig.savefig(fig_path)
            except Exception as e:
                sys.exit('Error! Unable to generate plots for visualization: %s. This may be due to unsuccessful installation of matplotlib package. For more, please see the installation note in the README file at `https://github.com/compdyn/ProSRS`.' % str(e))




"""
Copyright (C) 2016-2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

Define constants.
"""
# file templates
STATE_NPZ_FILE_TEMP = 'optimizer_state_%s.npz' # file that saves optimizer state (%s is optimization problem name)
STATE_PKL_FILE_TEMP = 'optimizer_state_%s.pkl' # file that saves optimizer state (%s is optimization problem name)
STATE_NPZ_TEMP_FILE_TEMP = 'optimizer_state_%s.temp.npz' # temporary file that saves optimizer state (%s is optimization problem name)
STATE_PKL_TEMP_FILE_TEMP = 'optimizer_state_%s.temp.pkl' # temporary file that saves optimizer state (%s is optimization problem name)


"""
Copyright (C) 2016-2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

Define utility functions.
"""
import numpy as np
from pathos.multiprocessing import ProcessingPool as Pool
from timeit import default_timer
from bisect import bisect_left


def eval_func(func, pts, n_proc=1, seeds=None, seed_func=None, save_files=None):
    """
    Evaluate a function in serial/parallel.
    
    Here we assume that parallelism takes place across multiple cores of a machine,
    and the Python package `pathos <https://pypi.org/project/pathos/0.2.0>` has 
    been installed.
    
    Args:
        
        func (callable): Function to be evaluated, of which the input is 1d array,
            the output is a scalar.
        
        pts (2d array): Points to be evaluated. Each point is one row of `pts`.
        
        n_proc (int, optional): Number of processes used for evaluation. If
            = 1, we evaluate `pts` in serial.
            
        seeds (list or None, optional): Random seed for each point in `pts`.
            If None, then no random seed will be set for any evaluation.
        
        seed_func (callable or None, optional): User-specified function for setting the random seed.
            If None, then we use ``numpy.random.seed`` method to set random seed.
            If callable, it is a function taking ``seed`` as an argument.
            Note: using ``numpy.random.seed`` may not always gaurantee
            the reproducibility. Here we provide an option for users to specify their own routines.
        
        save_files (list or None): Specify ``.npz`` file where each evaluation will be saved.
            If None, then no evaluation will be saved. It may be helpful to save evaluations
            for debugging purposes.
    
    Returns:
        
        vals (1d array): Function values at the points `pts`.
    
    Raises:
        
        ValueError: If n_proc is less than 1.
    """ 
    def eval_wrapper(arg):
        """
        A wrapper for evaluating the function `func`.
        
        Args:
            
            arg (tuple): Function argument = (pt, seed, save_file) with
            
                pt (1d array): A point to be evaluated.
            
                seed (int or None): Random seed for the evaluation. 
                    If None, then no random seed will be set for the evaluation.
            
                save_file (str or None): Specify ``.npz`` file where the evaluation will be saved.
                    If None, then the evaluation will not be saved.
        
        Returns:
            
            val (float): Function value at the point `pt`.
        """
        # parse argument
        pt, seed, save_file = arg
        
        if seed is not None:
            # set random seed
            if seed_func is None:
                np.random.seed(seed)
            else:
                assert(callable(seed_func)), 'seed_func is not callable.'
                seed_func(seed)
        
        t1 = default_timer()
        val = func(pt)
        t2 = default_timer()
        eval_time = t2-t1 # elapsed time for the evaluation
        
        if save_file is not None:
            np.savez(save_file, pt=pt, val=val, seed=seed, eval_time=eval_time)
        
        return val
        
    n_pts = pts.shape[0]
    seeds = [None]*n_pts if seeds is None else seeds
    save_files = [None]*n_pts if save_files is None else save_files
    # sanity check
    assert(len(seeds) == len(save_files) == n_pts)
    assert(callable(func)), 'func is not callable.'
    
    if n_proc == 1:
        vals = np.array([eval_wrapper(x) for x in zip(pts, seeds, save_files)])
    elif n_proc > 1:
        assert(type(n_proc) is int)
        pool = Pool(nodes=n_proc)
        vals = np.array(pool.map(eval_wrapper, list(zip(pts, seeds, save_files))))
    else:
        raise ValueError('Invalid n_proc value.')
    # sanity check
    assert(vals.ndim == 1), 'vals should be 1d numpy array.'
       
    return vals


def unique_row(array):
    """
    Sort a 2d array row-wise and get unique rows of it.
    
    Note: experiments show that this function produces results faster than ``np.unique`` method,
    especially when the number of rows of the array is large.
    
    Args:
        
        array (2d array): Array to get unique rows from
    
    Returns:
    
        unique_arr (2d array): Sorted array of unique rows. If `array` is empty (i.e.,
            ``array.shape = (0,0)``), then the empty `array` will be returned.
    """
    # sanity check
    assert(array.ndim == 2)
    
    if array.size > 0: # i.e. non-empty        
        
        sorted_arr = array[np.lexsort(array.T), :]
        # get unique rows
        row_ix = np.append([True], np.any(np.diff(sorted_arr, axis=0), axis=1))
        unique_arr = sorted_arr[row_ix]
    else:
        unique_arr = array.copy()
        
    return unique_arr

 
def put_back_box(pt, domain):
    """
    Put points back to a box-shaped domain, if there are any outside the domain.
    Any outside points are relocated to their nearest points in the domain.
    
    Args:
        
        pt (2d array): Points to be put back.
        
        domain (list of tuples): Box-shaped domain.
                For example, `domain` = [(0, 1), (0, 2)] means defining a 2D domain: 
                with first coordinate in [0, 1] and second coordinate in [0, 2].
        
    Returns:
        
        uniq_pt (2d array): Unique points after being put back (any duplicated points are squashed).
        
        raw_pt (2d array): Original points after being put back (with possible duplicate points).
    """
    n_pt, dim = pt.shape
    assert(len(domain)==dim)
    
    for j in range(dim):
        min_bd = domain[j][0]
        max_bd = domain[j][1]
        assert(max_bd > min_bd)
        if j == 0:
            ind = np.logical_and(pt[:, j] > min_bd, pt[:, j] < max_bd) # indicator that shows if a number is inside the range
        else:
            ind = np.logical_or(ind, np.logical_and(pt[:, j] > min_bd, pt[:, j] < max_bd))
        pt[:, j] = np.maximum(min_bd, np.minimum(max_bd, pt[:, j]))       
    pt_1 = pt[ind]
    pt_2 = pt[np.logical_not(ind)]
    raw_pt = np.vstack((pt_1, pt_2))
    pt_2 = unique_row(pt_2) # find unique points
    uniq_pt = np.vstack((pt_1, pt_2))
    
    return uniq_pt, raw_pt


def scale_zero_one(arr):
    """
    Scale an array to the range [0, 1] with minimum corresponding to 0 
    and maximum to 1.
    
    Args:
        
        arr (array): Array to be scaled.
        
    Returns:
        
        scale_arr (array): Scaled array (same shape as `arr`).
    """
    max_arr, min_arr = np.amax(arr), np.amin(arr)
    if max_arr != min_arr:
        scale_arr = (arr-min_arr)/(max_arr-min_arr)
    else:
        scale_arr = np.ones_like(arr)
        
    return scale_arr


def scale_one_zero(arr):
    """
    Scale an array to the range [0, 1] with minimum corresponding to 1 
    and maximum to 0.
    
    Args:
        
        arr (array): Array to be scaled.
        
    Returns:
        
        scale_arr (array): Scaled array (same shape as `arr`).
    """         
    max_arr, min_arr = np.amax(arr), np.amin(arr)
    if max_arr != min_arr:
        scale_arr = (max_arr-arr)/(max_arr-min_arr)
    else:
        scale_arr = np.ones_like(arr)
        
    return scale_arr


def eff_npt(pt, domain):
    """
    Get effective number of points within in a box-shaped domain.
    The effective number ``n_{eff}`` is computed by first uniformly partitioning the domain
    into cells with the number of cells per dimension being equal to ``ceil(n^{1/d})``, 
    where ``n`` is the number of points. Then ``n_{eff}`` is the number of cells that 
    are occupied by at least one point.
    
    Args:
        
        pt (2d array): Points in the domain.
        
        domain (list of tuples): Box-shaped domain.
            For example, `domain` = [(0, 1), (0, 2)] means defining a 2D domain: 
            with first coordinate in [0, 1] and second coordinate in [0, 2].
                
    Returns:
        
        n_eff (int): Effective number of points.
    """
    npt, dim = pt.shape
    assert(dim == len(domain))
    m = int(np.ceil(npt**(1./dim)))
    assert(m>1)
    # find grid for each axis
    grid_arr = np.zeros((dim, m+1))
    for j, axis_bd in enumerate(domain):
        grid_arr[j] = np.linspace(axis_bd[0], axis_bd[1], num=m+1)
    # find location of each sample (save in string)
    loc_samp = ['']*npt
    for i in range(npt):
        for j, grid in enumerate(grid_arr):
            assert(grid[0] <= pt[i, j] <=grid[-1])
            ix = bisect_left(grid, pt[i, j])
            assert(ix < m+1)
            loc = 0 if ix == 0 else ix-1
            loc_samp[i] += str(loc)
    # find number of samples with unique location
    n_eff = len(set(loc_samp))        
    assert(n_eff <= npt)
    
    return n_eff


def boxify(pt, domain):
    """
    Check whether each point is in the box-shaped domain.

    Args:
        
        pt (2d array): Points.
        
        domain (list of tuples): Box-shaped domain.
            For example, `domain` = [(0, 1), (0, 2)] means defining a 2D domain: 
            with first coordinate in [0, 1] and second coordinate in [0, 2].
    Returns:
        
        in_box_ix (1d array of bool): Indicate whether each point in `pt` is in the box.
        
        out_box_ix (1d array of bool): Indicate whether each point in `pt` is out of the box.
    """
    npt, dim = pt.shape
    for i in range(dim):
        assert(domain[i][0] < domain[i][1])
        ix = np.logical_and(domain[i][0] <= pt[:, i], domain[i][1] >= pt[:, i])
        if i == 0:
            in_box_ix = ix # indicates whether the points are in the box
        else:
            in_box_ix = np.logical_and(in_box_ix,ix)
    out_box_ix = np.logical_not(in_box_ix)
    
    return in_box_ix, out_box_ix


def domain_intersect(domain1, domain2):
    """
    Find intersection of two domains.
    
    Args:
        
        domain1 (list of tuples): First box-shaped domain.
            For example, `domain1` = [(0, 1), (0, 2)] means defining a 2D domain: 
            with first coordinate in [0, 1] and second coordinate in [0, 2].
        
        domain2 (list of tuples): Second box-shaped domain.
            For example, `domain2` = [(0, 1), (0, 2)] means defining a 2D domain: 
            with first coordinate in [0, 1] and second coordinate in [0, 2].
            
    Returns:
        
        inter_domain (list of tuples or None): Intersection of `domain1` and `domain2`. 
            If None, then the intersection is empty.
    """
    lb1, ub1 = zip(*domain1)
    lb2, ub2 = zip(*domain2)
    lb1, ub1 = np.array(lb1), np.array(ub1)
    lb2, ub2 = np.array(lb2), np.array(ub2)
    assert(np.all(ub1>=lb1) and np.all(ub2>=lb2))
    assert(lb1.shape==ub1.shape==lb2.shape==ub2.shape)
    inter_lb = np.maximum(lb1,lb2)
    inter_ub = np.minimum(ub1,ub2)
    if np.all(inter_ub>inter_lb):
        inter_domain = list(zip(inter_lb, inter_ub))
    else:
        inter_domain = None
        
    return inter_domain


"""
Copyright (C) 2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

Define release version.
"""
__version__ = '0.1.4'


"""
Copyright (C) 2016-2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

Define utility classes.
"""
import sys


class std_out_logger(object):
    """
    A class that handles logging standard outputs to a file.
    
    Usage:
        
        sys.stdout = std_out_logger(log_file) # redefine behavior of ``sys.stdout``
    """
    def __init__(self, log_file):
        """
        Args:
            
            log_file (str): File that standard outputs will be written to.
        """
        self.terminal = sys.stdout
        self.log = open(log_file, 'a')

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)
    
    def flush(self): # needed for Python 3 compatibility.
        pass


class std_err_logger(object):
    """
    A class that handles logging standard errors to a file.
    
    Usage:
        
        sys.stderr = std_err_logger(log_file) # redefine behavior of ``sys.stderr``
    """
    def __init__(self, log_file):
        """
        Args:
            
            log_file (str): File that standard errors will be written to.
        """
        self.terminal = sys.stderr
        self.log = open(log_file, 'a')

    def write(self, message):
        self.terminal.write(message)
        self.log.write(message)
    
    def flush(self): # needed for Python 3 compatibility.
        pass


"""
Copyright (C) 2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

"""
import setuptools

# read the long description of the package from README
with open("README.md", 'r') as f:
    long_description = f.read()
    
# read __version__ from the file
exec(open("prosrs/version.py").read())

setuptools.setup(
    name="prosrs",
    version=__version__,
    description="A tree-based parallel surrogate optimization algorithm for optimizing noisy expensive functions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/compdyn/ProSRS",
    author="Chenchao Shou",
    author_email="cshou3@illinois.edu",
    license="University of Illinois/NCSA Open Source",
    classifiers=[
            "Programming Language :: Python :: 2",
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Natural Language :: English",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX :: Linux",
            "License :: OSI Approved :: University of Illinois/NCSA Open Source License",
            "Intended Audience :: Developers",
            "Intended Audience :: Science/Research",
            "Topic :: Scientific/Engineering",
            "Topic :: Software Development"
    ],
    keywords="optimization algorithm",
    packages=setuptools.find_packages(exclude=['test']),
    install_requires=['numpy', 'scipy', 'matplotlib', 'pyDOE', 'pathos', 'scikit-learn']      
)


"""
Copyright (C) 2016-2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

Contain optimization benchmark problems.
"""
import numpy as np
from .problem import Problem


def benchmark(func_name, domain=None, sd=None):
    """
    Generate an optimization benchmark problem.
    
    Args:
        
        func_name (str): Benchmark function name.
        
        domain (list of tuples or None, optional): Optimization domain.
                For example, `domain` = [(0, 1), (0, 2)] means optimizing a
                2D function of argument ``x`` on the domain: 
                ``0 <= x[0] <= 1`` and ``0 <= x[1] <= 2``. If None, we use default
                value associated with `func_name`.
                
        sd (float, optional): Standard deviation of Gaussian noise added to the benchmark function.
            If None, then we use the default value associated with `func_name`.
    
    Returns:
        
        prob (`Problem` type): Optimization problem instance for benchmark `func_name`.
        
    Raises:
        
        ValueError: If `func_name` is not implemented.
    """
    if func_name == 'Schaffer':
        sd = 0.02 if sd is None else sd
        prob = benchmark_to_problem(Schaffer(domain), sd)
        
    elif func_name == 'Goldstein':
        sd = 2. if sd is None else sd
        prob = benchmark_to_problem(Goldstein(domain), sd)
        
    elif func_name == 'DropWave':
        sd = 0.02 if sd is None else sd
        prob = benchmark_to_problem(DropWave(domain), sd)
        
    elif func_name == 'SixHumpCamel':
        sd = 0.1 if sd is None else sd
        prob = benchmark_to_problem(SixHumpCamel(domain), sd)
        
    elif func_name == 'Ackley10':
        sd = 1. if sd is None else sd
        prob = benchmark_to_problem(Ackley(10, domain), sd)
        
    elif func_name == 'Alpine10':
        sd = 1. if sd is None else sd
        prob = benchmark_to_problem(Alpine(10, domain), sd)
        
    elif func_name == 'Hartmann6':
        sd = 0.05 if sd is None else sd
        prob = benchmark_to_problem(Hartmann6(domain), sd)
        
    elif func_name == 'Rastrigin2':
        sd = 0.5 if sd is None else sd
        prob = benchmark_to_problem(Rastrigin(2, domain), sd)
        
    elif func_name == 'Griewank10':
        sd = 2. if sd is None else sd
        prob = benchmark_to_problem(Griewank(10, domain), sd)
        
    elif func_name == 'Levy10':
        sd = 1. if sd is None else sd
        prob = benchmark_to_problem(Levy(10, domain), sd)
        
    elif func_name == 'SumPower10':
        sd = 0.05 if sd is None else sd
        prob = benchmark_to_problem(SumPower(10, domain), sd)
        
    elif func_name == 'PowerSum':
        sd = 1. if sd is None else sd
        prob = benchmark_to_problem(PowerSum(domain), sd)
        
    else:
        raise ValueError('Unknown func_name.') 
    
    return prob


def benchmark_to_problem(benchmark, sd):
    """
    Convert a benchmark instance to its corresponding optimization problem instance.
    
    Args:
        
        benchmark (Benchmark function type (e.g. `Schaffer`)): Benchmark instance.
        
        sd (float): Standard deviation of Gaussian noise added to the benchmark function.
        
    Returns:
        
        prob (`Problem` type): Optimization problem instance corresponding to `benchmark`.
    
    Raises:
        
        ValueError: If `sd` < 0.
    
    """
    if sd < 0:
        raise ValueError('sd must be non-negative.')
        
    func = lambda x: benchmark.f(x)+np.random.normal(0, sd) # add Gaussian noise
    prob = Problem(func, domain=benchmark.domain, name=benchmark.name, 
                   true_func=benchmark.f, min_loc=benchmark.min, 
                   min_true_func=benchmark.fmin, noise_type='Gaussian', sd=sd)
    
    return prob


class Schaffer:
    """
    `Schaffer N.2 function <https://www.sfu.ca/~ssurjano/schaffer2.html>`.
    """
    def __init__(self, domain=None):
        """
        Constructor.
        
        Args:
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.domain = [(-100., 100.), (-100., 100.)] if domain is None else domain
        self.dim = len(self.domain)
        self.min = np.array([(0., 0.)]) # global minimum locations
        self.fmin = 0. # global minimum
        self.name = 'Schaffer'
        
    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        x1 = x[0]
        x2 = x[1]
        y = 0.5+(np.sin(x1**2-x2**2)**2-0.5)/(1+(1e-3)*(x1**2+x2**2))**2
        
        return y
        
    
class Goldstein:
    """
    `Goldstein-Price function <https://www.sfu.ca/~ssurjano/goldpr.html>`.
    """
    def __init__(self, domain=None):
        """
        Constructor.
        
        Args:
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.domain = [(-2., 2.),(-2., 2.)] if domain is None else domain
        self.dim = len(self.domain)
        self.min = np.array([(0., -1.)]) # global minimum locations
        self.fmin = 3. # global minimum
        self.name = 'Goldstein'
    
    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        x1 = x[0]
        x2 = x[1]
        fact1a = (x1 + x2 + 1)**2
        fact1b = 19 - 14*x1 + 3*x1**2 - 14*x2 + 6*x1*x2 + 3*x2**2
        fact1 = 1 + fact1a*fact1b
        fact2a = (2*x1 - 3*x2)**2
        fact2b = 18 - 32*x1 + 12*x1**2 + 48*x2 - 36*x1*x2 + 27*x2**2
        fact2 = 30 + fact2a*fact2b
        y = fact1*fact2
        
        return y


class SixHumpCamel:
    """
    `Six-hump camel function <https://www.sfu.ca/~ssurjano/camel6.html>`.
    """
    def __init__(self, domain=None):
        """
        Constructor.
        
        Args:
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.domain = [(-3., 3.), (-2., 2.)] if domain is None else domain
        self.dim = len(self.domain)
        self.min = np.array([(0.0898, -0.7126), (-0.0898, 0.7126)]) # global minimum locations
        self.fmin = -1.0316 # global minimum
        self.name = 'SixHumpCamel'

    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        x1 = x[0]
        x2 = x[1]
        term1 = (4-2.1*x1**2 + (x1**4)/3) * x1**2
        term2 = x1*x2
        term3 = (-4+4*x2**2) * x2**2
        y = term1 + term2 + term3
        
        return y


class DropWave:
    """
    `Drop-wave function <https://www.sfu.ca/~ssurjano/drop.html>`.
    """
    def __init__(self, domain=None):
        """
        Constructor.
        
        Args:
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.domain = [(-5.12, 5.12), (-5.12, 5.12)] if domain is None else domain
        self.dim = len(self.domain)
        self.min = np.array([(0., 0.)]) # global minimum locations
        self.fmin = -1. # global minimum
        self.name = 'DropWave'

    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        y = -(1+np.cos(12*np.sqrt(x[0]**2+x[1]**2))) / (0.5*(x[0]**2+x[1]**2)+2)
        
        return y

        
class Alpine:
    """
    `Alpine function <http://infinity77.net/global_optimization/test_functions_nd_A.html#go_benchmark.Alpine01>`.
    """
    def __init__(self, dim, domain=None):
        """
        Constructor.
        
        Args:
            
            dim (int): Dimension of function.
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.dim = dim
        self.domain = [(-10., 10.)]*dim if domain is None else domain
        self.min = np.array([tuple([0.]*dim)]) # global minimum locations
        self.fmin = 0. # global minimum
        self.name = 'Alpine%d' % dim
    
    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        y = np.absolute(x*np.sin(x) + 0.1*x).sum()
        
        return y


class Ackley:
    """
    `Ackley function <https://www.sfu.ca/~ssurjano/ackley.html>`.
    """
    def __init__(self, dim, domain=None):
        """
        Constructor.
        
        Args:
            
            dim (int): Dimension of function.
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.dim = dim
        self.domain =[(-32.768, 32.768)]*dim if domain is None else domain
        self.min = np.array([tuple([0.]*dim)]) # global minimum locations
        self.fmin = 0. # global minimum
        self.name = 'Ackley%d' % dim

    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        y = (20+np.exp(1)-20*np.exp(-0.2*np.sqrt((x**2).sum()/self.dim))\
            -np.exp(np.cos(2*np.pi*x).sum()/self.dim))
        
        return y

        
class Hartmann6:
    """
    `Hartmann6 function <https://www.sfu.ca/~ssurjano/hart6.html>`.
    """
    def __init__(self, domain=None):
        """
        Constructor.
        
        Args:
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.dim = 6
        self.domain =[(0., 1.)]*self.dim if domain is None else domain
        self.min = np.array([(0.20169, 0.150011, 0.476874, 0.275332, 0.311652, 0.6573)]) # global minimum locations
        self.fmin = -3.32237 # global minimum
        self.name = 'Hartmann6'
        
    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        a = np.array([1, 1.2, 3, 3.2])
        A = np.array([[10, 3, 17, 3.5, 1.7, 8], [0.05, 10, 17, 0.1, 8, 14], 
                      [3, 3.5, 1.7, 10, 17, 8], [17, 8, 0.05, 10, 0.1, 14]])     
        P = 10**(-4)*np.array([[1312, 1696, 5569, 124, 8283, 5886], 
                [2329, 4135, 8307, 3736, 1004, 9991], [2348, 1451, 3522, 2883, 3047, 6650],
                [4047, 8828, 8732, 5743, 1091, 381]])
        x = x.reshape((1, -1))
        y = np.zeros((1, 1))
        for i in range(4):
            ai, Ai, Pi = a[i], A[i], P[i]
            y += -ai*np.exp(-np.dot((x-np.outer(np.ones(1), Pi))**2, Ai.reshape((-1,1))))
        y = y.item(0)
        
        return y

        
class Rastrigin:
    """
    `Rastrigin function <https://www.sfu.ca/~ssurjano/rastr.html>`.
    """
    def __init__(self, dim, domain=None):
        """
        Constructor.
        
        Args:
            
            dim (int): Dimension of function.
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.dim = dim
        self.domain =[(-5.12, 5.12)]*dim if domain is None else domain
        self.min = np.array([tuple([0.]*dim)]) # global minimum locations
        self.fmin = 0. # global minimum
        self.name = 'Rastrigin%d' % dim
        
    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        y = 10*self.dim+np.sum(x**2-10*np.cos(2*np.pi*x))
        
        return y


class Griewank:
    """
    `Griewank function <https://www.sfu.ca/~ssurjano/griewank.html>`.
    """
    def __init__(self, dim, domain=None):
        """
        Constructor.
        
        Args:
            
            dim (int): Dimension of function.
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.dim = dim
        self.domain =[(-600., 600.)]*dim if domain is None else domain
        self.min = np.array([tuple([0.]*dim)]) # global minimum locations
        self.fmin = 0. # global minimum
        self.name = 'Griewank%d' % dim
        
    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        x = x.reshape((1, -1))
        y = np.sum(x**2, axis=1)/4000.-np.prod(np.cos(x/np.sqrt(np.outer(np.ones(1), range(1, self.dim+1)))), axis=1)+1
        y = y.item(0)
        
        return y


class Levy:
    """
    `Levy function <https://www.sfu.ca/~ssurjano/levy.html>`.
    """
    def __init__(self, dim, domain=None):
        """
        Constructor.
        
        Args:
            
            dim (int): Dimension of function.
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.dim = dim
        self.domain =[(-10., 10.)]*dim if domain is None else domain
        self.min = np.array([tuple([1.]*dim)]) # global minimum locations
        self.fmin = 0. # global minimum
        self.name = 'Levy%d' % dim
        
    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        W = 1+(x-1)/4.
        W0 = W[0]
        W1 = W[:-1]
        W2 = W[-1]
        term1 = np.sin(np.pi*W0)**2
        term2 = np.sum((W1-1)**2*(1+10*np.sin(np.pi*W1+1)**2))
        term3 = (W2-1)**2*(1+np.sin(2*np.pi*W2)**2)
        y = term1+term2+term3
        
        return y


class SumPower:
    """
    `Sum of powers function <https://www.sfu.ca/~ssurjano/sumpow.html>`.
    """
    def __init__(self, dim, domain=None):
        """
        Constructor.
        
        Args:
            
            dim (int): Dimension of function.
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.dim = dim
        self.domain =[(-1., 1.)]*dim if domain is None else domain
        self.min = np.array([tuple([0.]*dim)]) # global minimum locations
        self.fmin = 0. # global minimum
        self.name = 'SumPower%d' % dim
        
    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        y = np.sum(np.absolute(x)**np.arange(2, self.dim+2))
        
        return y


class PowerSum:
    """
    `Power sum function <https://www.sfu.ca/~ssurjano/powersum.html>`.
    """
    def __init__(self, domain=None):
        """
        Constructor.
        
        Args:
            
            domain (list of tuples or None, optional): Optimization domain.
                If None, we use the default value.
                
        """
        self.dim = 4
        self.domain =[(0., 4.)]*self.dim if domain is None else domain
        self.min = np.array([(1., 2., 2., 3.)]) # global minimum locations
        self.fmin = 0. # global minimum
        self.name = 'PowerSum'
        
    def f(self, x):
        """
        Expression of function.
        
        Args:
            
            x (1d array): Function input.
                
        Returns:
            
            y (float): Function evaluation at `x`.
            
        Raises:
            
            ValueError: If dimension of `x` is wrong.
            
        """
        # sanity check
        if x.shape != (self.dim, ):
            raise ValueError('Wrong dimension for x.')
        # evaluate the function
        y = 0.
        bval = [8., 18., 44., 114.]
        for i,b in enumerate(bval):
            y += (np.sum(x**(i+1))-b)**2
        
        return y



"""
Copyright (C) 2016-2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

Implement ProSRS algorithm.
"""
import numpy as np
import os, sys, pickle, shutil, warnings
import multiprocessing as mp
from pathos.multiprocessing import ProcessingPool as Pool
from timeit import default_timer
from pyDOE import lhs
from scipy.spatial.distance import cdist
from ..utility.constants import STATE_NPZ_FILE_TEMP, STATE_PKL_FILE_TEMP, STATE_NPZ_TEMP_FILE_TEMP, STATE_PKL_TEMP_FILE_TEMP
from ..utility.classes import std_out_logger, std_err_logger
from ..utility.functions import eval_func, put_back_box, scale_one_zero, scale_zero_one, eff_npt, boxify, domain_intersect
from .surrogate import RBF_reg
try:
    import matplotlib.pyplot as plt
    from matplotlib.ticker import MaxNLocator
except:
    pass


class Optimizer:
    """
    A class that handles optimization using ProSRS algorithm.
    """
    def __init__(self, prob, n_worker, n_iter=None, n_iter_doe=None, n_cycle=2, resume=False, 
                 seed=1, seed_func=None, parallel_training=False, out_dir='out'):
        """
        Constructor.
        
        Args:
            
            prob (`prosrs.Problem` type): Optimization problem.
            
            n_worker (int): Number of workers for the optimization.
                This also determines the number of proposed or evaluated
                points per iteration.
                
            n_iter (int or None, optional): Total number of iterations for the optimization.
                If ``int``, the optimization will terminate upon finishing running `n_iter` iterations. 
                If None, then we use number of optimization cycles `n_cycle` as the termination condition.
            
            n_iter_doe (int or None, optional): Number of iterations for DOE.
                If None, then we use default value.
                
            n_cycle (int, optional): Total number of optimization sycles.
                This parameter takes effect only when `n_iter` is None, and 
                the optimization will terminate upon finishing `n_cycle` cycles.
            
            resume (bool, optional): Whether to resume from the last run.
                The information of the last run will be read from the directory `out_dir`.
            
            seed (int or None, optional): Random seed for the optimizer.
                If None, then we do not set random seed for the optimization.
            
            seed_func (callable or None, optional): User-specified function for setting the random seed for evaluations.
                If None, then we use ``numpy.random.seed(seed)`` method to set random seed.
                If callable, it is a function taking ``seed`` as an argument.
                Note: using ``numpy.random.seed`` may not always gaurantee
                the reproducibility. Here we provide an option for users to specify their own routines.
            
            parallel_training (bool, optional): Whether to train a RBF surrogate in parallel.
                If True, then we will use all the available processes (cores) during training.
                We use `pathos.multiprocessing` module for parallelism. Our tests have
                shown that depending on the machine and the optimization function,
                sometimes parallelism may cause memory issues. So we disable it by default.
                
            out_dir (str, optional): Output directory.
                All the output files (e.g., optimization status file) will be saved to
                `out_dir` directory.
        """
        # class members (convention: variables wih prefix '_' are constant parameters during optimization).
        self._prob = prob # optimization problem
        self._dim = prob.dim # dimention of optimization problem.
        self._n_worker = n_worker # number of workers.
        self._n_iter = n_iter # total number of iterations.
        self._n_cycle = n_cycle # total number of optimization cycles.
        self._resume = resume # whether to resume from the last run.
        self._seed = seed # random seed for the optimizer.
        self._seed_func = seed_func # function for setting random seed for evaluations.
        self._parallel_training = parallel_training # whether to use parallelism for training RBF models.
        self._out_dir = out_dir # output directory.
        self._n_cand_fact = 1000 # number of candidate = self._n_cand_fact * self._dim.
        self._wgt_pat_bd = [0.3, 1.] # the bound of the weight pattern in the SRS method.
        self._normalize_data = True # whether to normalize data when training RBF regression model.
        self._init_gamma = 0. # initial weight exponent in the SRS method (>=0). If zero, then disable weighting in RBF regression.
        self._delta_gamma = 2. # amount of decrease of weight exponent whenever failure occurs.
        self._init_sigma = 0.1 # initial sigma value in the SRS method (controls initial spread of Type II candidate points).
        self._max_n_reduce_sigma = 2 # max number of times of halving self.sigma before zooming in/restart. Critical sigma value = self._init_sigma * 0.5**self._max_n_reduce_sigma.
        self._rho = 0.4 # zoom-in factor. Must be in (0, 1).
        self._init_p = 1. # initial p value in the SRS method (controls initial proportion of Type I candidate points).
        self._init_beta = 0.02 # initial beta value (= initial probability of zooming out).
        self._min_beta = 0.01 # minimum beta value (= minimum probability of zooming out).
        self._alpha = 1. # parameter controlling decreasing rate of p value: p = p*n_{eff}**(-self._alpha/self.dim).
        self._lambda_range = [1e-7, 1e4] # range of regularization parameter in RBF regression, over which we determine regularization constant in L2 regularization.
        self._rbf_kernel = 'multiquadric' # RBF kernel. See `scipy.interpolate.Rbf <https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.interpolate.Rbf.html>`.
        self._rbf_poly_deg = 0 # degree of RBF polynomial tail (either 0 or 1). If zero, then no polynomial tail.
        self._n_fold = 5 # number of folds for the cross validation in training a RBF model.
        self._resol = 0.01 # resolution parameter for determining whether to restart.
        self._use_eff_npt = True # whether to use effective number of points for the dynamics of p value.
        self._max_C_fail = max(2, int(np.ceil(self._dim/float(self._n_worker)))) # maximum number of consecutive failures before halving self.sigma value.
        self._n_iter_doe = int(np.ceil(3/float(self._n_worker))) if n_iter_doe is None else n_iter_doe # number of iterations for DOE.
        self._n_iter_doe = min(self._n_iter, self._n_iter_doe) if self._n_iter is not None else self._n_iter_doe # adjusted for the fact that self._n_iter_doe <= self._n_iter.
        self._n_doe_samp = self._n_iter_doe * self._n_worker # number of DOE samples
        self._n_cand = self._n_cand_fact * self._dim # number of candidate points in SRS method
        self._pool_rbf = Pool(nodes=mp.cpu_count()) if self._parallel_training else None # parallel pool used for RBF regression
        self._state_npz_file = os.path.join(self._out_dir, STATE_NPZ_FILE_TEMP % self._prob.name) # file that saves optimizer state (needed for resume)
        self._state_pkl_file = os.path.join(self._out_dir, STATE_PKL_FILE_TEMP % self._prob.name) # file that saves optimizer state (needed for resume)
        self._state_npz_lock_file = self._state_npz_file+'.lock' # a lock file that may be generated in some system, which prevents reading data from `self._state_npz_file`.
        self._state_pkl_lock_file = self._state_pkl_file+'.lock' # a lock file that may be generated in some system, which prevents reading data from `self._state_pkl_file`.
        self._state_npz_temp_file = os.path.join(self._out_dir, STATE_NPZ_TEMP_FILE_TEMP % self._prob.name) # a temporary file that holds data for `self._state_npz_file`.
        self._state_pkl_temp_file = os.path.join(self._out_dir, STATE_PKL_TEMP_FILE_TEMP % self._prob.name) # a temporary file that holds data for `self._state_pkl_file`.
        self._verbose_dot_len = 10 # number of dots to display for verbose messages.
        
        # sanity check
        assert(type(self._n_worker) is int and self._n_worker > 0)
        assert(0 <= self._wgt_pat_bd[0] <= self._wgt_pat_bd[1] <= 1 and len(self._wgt_pat_bd) == 2)
        assert(self._delta_gamma >= 0)
        assert(type(self._max_n_reduce_sigma) is int and self._max_n_reduce_sigma >= 0)
        assert(0 < self._rho < 1)
        assert(0 <= self._init_p <= 1)
        assert(0 <= self._min_beta <= self._init_beta <= 1)
        assert(self._alpha > 0)
        assert(0 < self._lambda_range[0] <= self._lambda_range[1] and len(self._lambda_range) == 2)
        assert(type(self._rbf_poly_deg) is int and self._rbf_poly_deg in [0, 1])
        assert(type(self._n_fold) is int and self._n_fold > 1)
        assert(0 < self._resol < 1)
        assert(type(self._max_C_fail) is int and self._max_C_fail > 0)
        assert(type(self._n_iter_doe) is int and self._n_iter_doe > 0)
        assert(type(self._n_cand) is int and self._n_cand >= self._n_worker)
        if type(self._n_iter) is int:
            assert(self._n_iter >= 0)
        else:
            assert(self._n_iter is None), 'n_iter is either an integer or None.'
        assert(type(self._n_cycle) is int and self._n_cycle >= 0)
        
        # create output directory, if not existent
        if not os.path.isdir(self._out_dir):
            os.makedirs(self._out_dir)
        
        # initialize the state of the optimizer
        if not self._resume:
            np.random.seed(self._seed) # set random seed.
            self.i_iter = 0 # iteration index (how many iterations have been run)
            self.i_cycle = 0 # number of optimization cycles that have been completed
            self.doe_samp = self.doe() # DOE samples
            self.i_iter_doe = 0 # iteration index during DOE phase (how many DOE iterations have been run in current cycle)
            self.t_build_arr = np.zeros(0) # time of building a RBF model for each iteration. If an iteration is DOE, = 0.  
            self.t_srs_arr = np.zeros(0) # time of running SRS method for each iteration. If an iteration is DOE, = 0. 
            self.t_prop_arr = np.zeros(0) # time of proposing new points for each iteration.
            self.t_eval_arr = np.zeros(0) # time of evaluating proposed points for each iteration.
            self.t_update_arr = np.zeros(0) # time of updating the optimizer state for each iteration.
            self.gSRS_pct_arr = np.zeros(0) # pertentage of global SRS (= percentage of Type I candidates = floor(10*p_val)/10.) for each iteration.
            self.zoom_lv_arr = np.zeros(0) # zoom level at the time of proposing new points for each iteration.
            self.x_tree = np.zeros((0, self._dim)) # all the evaluated (proposed) points so far in a tree.
            self.y_tree = np.zeros(0) # (noisy) y values of `self.x_tree`.
            self.x_all = np.zeros((0, self._dim)) # all the evaluated (proposed) points so far in the course of optimization.
            self.y_all = np.zeros(0) # (noisy) y values of `self.x_all`.
            self.seed_all = np.zeros(0) # random seeds for points in `self.x_all`.
            self.best_x = np.ones(self._dim)*np.nan # best point so far.
            self.best_y = np.nan # (noisy) y value of the best point `self.best_x`.
            self.zoom_lv = 0 # zoom level (zero-based).
            self.act_node_ix = 0 # index of the activate node for the zoom level `self.zoom_lv` (zero-based).
            self.srs_wgt_pat = np.linspace(self._wgt_pat_bd[0], self._wgt_pat_bd[1], self._n_worker) # weight pattern in the SRS method.
            self.tree = self.init_tree() # initialize optimization tree
            self.eval_seeds = self._seed+1+np.arange(self.i_iter*self._n_worker, (self.i_iter+1)*self._n_worker, dtype=int) # random seeds for parallel evaluations
        else:
            # load optimizer state from the last run
            self.load_state()
        
    
    def show(self, select=['problem', 'config', 'status', 'result', 'post_result'], n_display=0):
        """
        Display the optimizer info.
        
        Args:
            
            select (list or tuple, optional): Select which info will be displayed.
                Possible values and their meanings:
                    'problem': optimization problem to be solved.
                    'config': optimization configuration.
                    'status': optimization status.
                    'result': optimization results.
                    'post_result': posterior evaluation results.
                    
            n_display (int, optional): Number of (top) points to be displayed for posterior evaluation results.
                This parameter takes effects only when `select` contains 'post_result'.
        """
        select_possible_vals = ['problem', 'config', 'status', 'result', 'post_result']
        assert(type(select) in [list, tuple] and set(select)<=set(select_possible_vals)), 'invalid select value'
        
        if 'problem' in select:
            print(self._prob)
            
        if 'config' in select:
            print('Optimization configuration:')
            print('- Number of workers: %d' % self._n_worker)
            if self._n_iter is None:
                print('- Termination criterion: stop after completing %d optimization cycles' % self._n_cycle)
            else:
                print('- Termination criterion: stop after completing %d iterations' % self._n_iter)
            is_resume = 'Yes' if self._resume else 'No'
            print('- Resuming from the last run?  %s' % is_resume)
            print("- Output directory: '%s'" % self._out_dir)
            
        if 'status' in select:
            print('Optimization status:')
            print('- Number of iterations completed: %d' % self.i_iter)
            print('- Number of cycles completed: %d' % self.i_cycle)
            print('- Zoom level of current node: %d' % self.zoom_lv) # note: zoom level of root node is zero
            node_domain = self.tree[self.zoom_lv][self.act_node_ix]['domain']
            print('- Domain of current node: %s' % ('{'+', '.join(["'%s': %s" % (x, str(tuple(v))) for x, v in \
                                                                   zip(self._prob.x_var, node_domain)])+'}'))

        if 'result' in select:
            print('Optimization result:')
            print('- Best point:')
            print('  '+', '.join(['%s = %g' % (x, v) for x, v in zip(self._prob.x_var, self.best_x)]))
            print('- Best (noisy) value:')
            print('  %s = %g' % (self._prob.y_var, self.best_y))
        
        if 'post_result' in select:
            try:
                n_pt = len(self.posterior_eval_x) # number of points evaluated in posterior runs
                assert(len(self.posterior_eval_y) == n_pt and type(self.i_iter_posterior_eval) is int)
                assert(type(n_display) is int and n_display >= 0)
                # display true best point and its value
                print('Posterior evaluation results:')
                print('- Condition: run ProSRS algorithm for %d iterations, then run posterior evaluations with %d Monte Carlo repeats'
                        % (self.i_iter_posterior_eval, self.posterior_eval_y.shape[1]))
                print('- True best point:')
                print('  '+', '.join(['%s = %g' % (x, v) for x, v in zip(self._prob.x_var, self.true_best_x)]))
                print('- True best value:')
                print('  %s = %g' % (self._prob.y_var, self.true_best_y))
                # display top points
                if n_display > 0:
                    # get (sorted) top points based on posterior mean values
                    if n_display > n_pt:
                        warnings.warn('Number of points in posterior evaluations is %d, less than n_display (%d). Reset n_display = %d.'
                                      % (n_pt, n_display, n_pt))
                        n_display = n_pt
                    sort_ix = np.argsort(self.posterior_mean)
                    display_x = self.posterior_eval_x[sort_ix][:n_display]
                    display_mean_y = self.posterior_mean[sort_ix][:n_display]
                    display_std_y = self.posterior_std[sort_ix][:n_display]
                    # format into strings
                    x_var_str = []
                    display_x_str = []
                    for j in range(self._dim):
                        var_len = len(self._prob.x_var[j])
                        x_str_list = ['%g' % x for x in display_x[:, j]]
                        str_len = max(var_len, max([len(x) for x in x_str_list]))
                        x_var_str.append(' '*(str_len-var_len)+self._prob.x_var[j]) # pad space at front
                        display_x_str.append([' '*(str_len-len(x))+x for x in x_str_list]) # pad space at front
                    display_x_str = [[x[j] for x in display_x_str] for j in range(n_display)]
                    y_str_list = ['%g' % x for x in display_mean_y]
                    y_var = 'mean of '+self._prob.y_var
                    str_len = max(len(y_var), max([len(x) for x in y_str_list]))
                    y_var_str = ' '*(str_len-len(y_var))+y_var # pad space at front
                    display_y_str = [' '*(str_len-len(x))+x for x in y_str_list] # pad space at front
                    y_err_str_list = ['%g' % x for x in display_std_y]
                    y_err_var = 'std of '+self._prob.y_var
                    str_len = max(len(y_err_var), max([len(x) for x in y_err_str_list]))
                    y_err_var_str = ' '*(str_len-len(y_err_var))+y_err_var # pad space at front
                    display_y_err_str = [' '*(str_len-len(x))+x for x in y_err_str_list] # pad space at front
                    print('- Top %d points sorted by mean response estimates in Column %d:' % (n_display, self._dim+1))
                    space = ' '*3
                    front_space = ' '*5
                    print(front_space+space.join(x_var_str+[y_var_str, y_err_var_str]))
                    for x, mean_y, std_y in zip(display_x_str, display_y_str, display_y_err_str):
                        print(front_space+space.join(x+[mean_y, std_y]))
            except:
                sys.exit("Error! No posterior evaluation results can be displayed. Please run 'posterior_eval' first.")
        

    def run(self, std_out_file=None, std_err_file=None, verbosity=1):
        """
        Run ProSRS algorithm.
        
        Args:
            
            std_out_file (str or None, optional): Standard output file path.
                If ``str``, then standard outputs will be directed to the file `std_out_file`.
                If None, then standard outputs will not be saved to a file.
            
            std_err_file (str or None, optional): Standard error file path.
                If ``str``, then standard errors will be directed to the file `std_err_file`.
                If None, then standard errors will not be saved to a file.
                
            verbosity (int, optional): Level of verbosity (0-2) while running the algorithm. 
                If zero, then no verbose. The larger this value is, the more information
                will be displayed.
        """  
        assert(0 <= verbosity <= 2)        
        # log standard outputs and standard errors.
        # here we write to a new file if we do not resume. Otherwise, we append to the old file.
        if std_out_file is not None:
            orig_std_out = sys.stdout
            if not self._resume:
                if os.path.isfile(std_out_file):
                    os.remove(std_out_file)
            sys.stdout = std_out_logger(std_out_file)    
        if std_err_file is not None:
            orig_std_err = sys.stderr
            if not self._resume:
                if os.path.isfile(std_err_file):
                    os.remove(std_err_file)
            sys.stderr = std_err_logger(std_err_file)
        
        if verbosity == 2:
            print('')
            self.show(select=['problem']) # show optimization problem
            print('')
            self.show(select=['config']) # show optimization configuration
        
        if verbosity > 0:
            print('\nStart running ProSRS algorithm ...')
            
        # main loop
        while not self.is_done():
            
            if verbosity > 0:
                print('\nIteration %d:' % (self.i_iter+1))
                
            # propose new points
            new_pt = self.propose(verbose=verbosity>0)        
            # evaluate proposed points
            new_val = self.eval_pt(new_pt, verbose=verbosity>0)
            # update optimizer state with the new evaluations
            self.update(new_pt, new_val, verbose=verbosity>0)
            
            if verbosity == 2:
                self.show(select=['status']) # show optimization status
            
            # flush standard outputs and standard errors to files
            if std_out_file is not None:
                sys.stdout.terminal.flush()
                sys.stdout.log.flush()
            if std_err_file is not None:
                sys.stderr.terminal.flush()
                sys.stderr.log.flush()
        
        if verbosity > 0:
            print('\nFinished running ProSRS algorithm.')
            
        if verbosity == 2:
            print('')
            self.show(select=['result']) # show optimization result
            
        # reset stdout and stderr
        if std_out_file is not None:
            sys.stdout = orig_std_out 
        if std_err_file is not None:
            sys.stderr = orig_std_err
        
        
    def doe(self, criterion='maximin'):
        """
        Design of experiments using Latin Hypercube Sampling (LHS).
        
        Args:
            
            criterion (str, optional): Sampling criterion for LHS.
                For details, see `pyDOE documentation <https://pythonhosted.org/pyDOE/randomized.html>`.
                
        Returns:
            
            samp (2d array): LHS samples. Each row is one sample.
        """
        unit_X = lhs(self._dim, samples=self._n_doe_samp, criterion=criterion) # unit_X: 2d array in unit cube
        samp = np.zeros_like(unit_X)
        for i in range(self._dim):
            samp[:, i] = unit_X[:, i]*(self._prob.domain[i][1]-self._prob.domain[i][0])+self._prob.domain[i][0] # scale and shift
        return samp
    
    
    def init_tree(self):
        """
        Initialize an optimization tree.
        
        Returns:
            
            tree (dict): Initial tree.
        """
        tree = {self.zoom_lv: [{'ix': np.arange(self._n_doe_samp, dtype=int), # indice of data for the tree node (w.r.t. `self.x_tree` or `self.y_tree`).
                                'domain': self._prob.domain, # domain of the tree node.
                                'parent_ix': None, # parent node index for the upper zoom level (zero-based). If None, there's no parent.
                                'beta': self._init_beta, # zoom-out probability.
                                'state': self.init_node_state() # state of the tree node.
                                }]}
        return tree
    
        
    def init_node_state(self):
        """
        Initialize the state of a node of the optimization tree.
        
        Returns:
            
            state (dict): Values of state variables.
        """
        
        state = {'p': self._init_p, # p value in the SRS method (controls proportion of Type I candidate points).
                 'Cr': 0, # counter that counts number of times of reducing the sigma value of local SRS.
                 'Cf': 0, # counter that counts number of consecutive failures.
                 'gamma': self._init_gamma # weight exponent parameter of weighted RBF
                 }
        
        return state
    
    
    def propose(self, verbose=True):
        """
        Propose new points for the next iteration.
        
        Args:
            
            verbose (bool, optional): Whether to verbose about proposing new points.
        
        Returns:
            
            new_pt (2d array): Proposed new points. Each row is one point.
        """
        tt1 = default_timer()
        
        if self.i_iter_doe < self._n_iter_doe:
            
            if verbose:
                sys.stdout.write('Proposing new points (DOE) '+'.'*self._verbose_dot_len)
            t1 = default_timer()
            
            # i.e., current iteration is in DOE phase
            new_pt = self.doe_samp[self.i_iter_doe*self._n_worker:(self.i_iter_doe+1)*self._n_worker]
            # the following variables will be saved in the `self.update` method
            # so we need to set them appropriately
            self.gSRS_pct = np.nan
            self.t_build = np.nan
            self.t_srs = np.nan
            
            t2 = default_timer()
            if verbose:
                sys.stdout.write(' Done (time took: %.2e sec).\n' % (t2-t1))
                
        else:
            # i.e., current iteration is in true optimization phase
            
            ########### Get activated node ##############
            
            self.act_node = self.tree[self.zoom_lv][self.act_node_ix]
            self.x_node = self.x_tree[self.act_node['ix']] # get X data of the node
            self.y_node = self.y_tree[self.act_node['ix']] # get Y data of the node
            self.p_val = self.act_node['state']['p']
            self.n_reduce_sigma = self.act_node['state']['Cr']
            self.n_fail = self.act_node['state']['Cf']
            self.gamma = self.act_node['state']['gamma']
            self.gSRS_pct = np.floor(10*self.p_val)/10. # pertentage of global SRS (= percentage of Type I candidates)
            self.sigma = self._init_sigma*0.5**self.n_reduce_sigma
            
            ########### Build RBF surrogate model ##############
            
            if verbose:
                sys.stdout.write('Building RBF regression model '+'.'*self._verbose_dot_len)
                
            t1 = default_timer()

            self.rbf_mod, _, _, _ = RBF_reg(self.x_node, self.y_node, self._lambda_range,
                                            normalize_data=self._normalize_data, wgt_expon=self.gamma,
                                            n_fold=self._n_fold, kernel=self._rbf_kernel, 
                                            poly_deg=self._rbf_poly_deg, pool=self._pool_rbf)                
            t2 = default_timer()
            self.t_build = t2-t1
            
            if verbose:
                sys.stdout.write(' Done (time took: %.2e sec).\n' % self.t_build)
            
            ########### Propose new points using SRS method ############## 
            
            if verbose:
                sys.stdout.write('Proposing new points '+'.'*self._verbose_dot_len)
                
            t1 = default_timer()
            
            new_pt = self.SRS()
            
            t2 = default_timer()
            self.t_srs = t2-t1
            
            if verbose:
                sys.stdout.write(' Done (time took: %.2e sec).\n' % self.t_srs)
                    
        tt2 = default_timer()
        self.t_prop = tt2-tt1
        
        return new_pt
        
    
    def SRS(self):
        """
        Propose new points using SRS method.
        
        Returns:
            
            new_pt (2d array): Proposed points.
        """ 
        # generate candidate points
        if self.gSRS_pct == 1:            
            # generate candidate points uniformly (global SRS)
            cand_pt = np.zeros((self._n_cand, self._dim))
            for d, bd in enumerate(self.act_node['domain']):
                cand_pt[:, d] = np.random.uniform(low=bd[0], high=bd[1], size=self._n_cand)        
        else:
            n_cand_gSRS = int(np.round(self._n_cand*self.gSRS_pct)) # number of candidate points for global SRS
            n_cand_lSRS = self._n_cand-n_cand_gSRS # number of candidate points for local SRS
            assert(n_cand_lSRS > 0) # sanity check
            # generate candidate points uniformly (global SRS)           
            cand_pt_gSRS = np.zeros((n_cand_gSRS, self._dim))
            if n_cand_gSRS > 0:
                for d, bd in enumerate(self.act_node['domain']):
                    cand_pt_gSRS[:, d] = np.random.uniform(low=bd[0], high=bd[1], size=n_cand_gSRS)
            # find x_star
            Y_fit = self.rbf_mod(self.x_node)
            min_ix = np.argmin(Y_fit)
            x_star = self.x_node[min_ix]
            assert(np.all([bd[0] <= x_star[j] <= bd[1] for j,bd in enumerate(self.act_node['domain'])])) # sanity check
            # find step size (i.e. std) for each coordinate of `x_star`
            step_size_arr = np.array([self.sigma*(bd[1]-bd[0]) for bd in self.act_node['domain']])
            assert(np.min(step_size_arr) > 0) # sanity check
            # generate candidate points (Gaussian about x_star, local SRS)
            cand_pt_lSRS = np.random.multivariate_normal(x_star, np.diag(step_size_arr**2), n_cand_lSRS)
            # combine two types of candidate points
            comb_cand_pt = np.vstack((cand_pt_gSRS, cand_pt_lSRS))
            # put candidate points back to the domain, if there's any outside
            uniq_cand_pt, raw_cand_pt = put_back_box(comb_cand_pt, self.act_node['domain'])            
            # get candidate points (``len(uniq_cand_pt) < n_worker`` is pathological case, almost never encountered in practice)
            cand_pt = uniq_cand_pt if len(uniq_cand_pt) >= self._n_worker else raw_cand_pt
        
        # select new points from candidate points
        n_cand = len(cand_pt)
        assert(n_cand >= self._n_worker)
        resp_cand = self.rbf_mod(cand_pt)
        resp_score = scale_zero_one(resp_cand) # response score
        # initializations
        new_pt = np.zeros((self._n_worker, self._dim))
        refer_pt = self.x_node.copy() # reference points based on which we compute distance scores
        # select points sequentially
        for j in range(self._n_worker):
            wt = self.srs_wgt_pat[j]
            if len(refer_pt) > 0:
                if j == 0:                                       
                    # distance matrix for `refer_pt` and `cand_pt`
                    dist_mat = cdist(cand_pt, refer_pt)   
                    dist_cand = np.amin(dist_mat, axis=1)                        
                else:
                    # distance to the previously proposed point
                    dist_prop_pt = cdist(cand_pt, new_pt[j-1].reshape((1, -1))).flatten()                    
                    dist_cand = np.minimum(dist_cand, dist_prop_pt)
                dist_score = scale_one_zero(dist_cand) # distance score
            else:
                # pathological case
                dist_score = np.zeros(n_cand) # distance score
            cand_score = resp_score*wt+(1-wt)*dist_score # candidate score        
            assert (np.amax(cand_score)<=1 and np.amin(cand_score)>=0) # sanity check            
            # select the best one based on the score
            min_ix = np.argmin(cand_score)
            new_pt[j] = cand_pt[min_ix]           
            # update variables
            refer_pt = np.vstack((refer_pt, new_pt[j].reshape((1, -1))))
            dist_cand = np.delete(dist_cand, min_ix)
            resp_score = np.delete(resp_score, min_ix)
            cand_pt = np.delete(cand_pt, min_ix, axis=0)
            n_cand -= 1
        
        return new_pt
        
    
    def eval_pt(self, x, verbose=True):
        """
        Evaluate proposed points.
        
        Args:
            
            x (2d array): Points to be evaluated. Each row is one point.
            
            verbose (bool, optional): Whether to verbose about the evaluation.
        
        Returns:
            
            y (1d array): Evaluations of points in `x`.
        """
        if verbose:
            sys.stdout.write('Evaluating proposed points '+'.'*self._verbose_dot_len)
            
        t1 = default_timer()
        
        y = eval_func(self._prob.f, x, n_proc=self._n_worker, seeds=self.eval_seeds.tolist(),
                      seed_func=self._seed_func)
        
        t2 = default_timer()
        self.t_eval = t2-t1
        
        if verbose:
            sys.stdout.write(' Done (time took: %.2e sec).\n' % self.t_eval)
        
        return y
    
    
    def update(self, new_x, new_y, verbose=True):
        """
        Update the state of the optimizer.
        
        Args:
            
            new_x (2d array): Proposed new points. Each row is one point.
            
            new_y (1d array): (Noisy) values of the points in `new_x`.
            
            verbose (bool, optional): Whether to verbose about updating the state of the optimizer.
        """
        if verbose:
            sys.stdout.write('Updating optimizer state '+'.'*self._verbose_dot_len)
            specific_msg = '' # specific message indicating the action
        t1 = default_timer()
        
        self.i_iter += 1
        self.eval_seeds = self._seed+1+np.arange(self.i_iter*self._n_worker, (self.i_iter+1)*self._n_worker, dtype=int)
        self.x_tree = np.vstack((self.x_tree, new_x))
        self.y_tree = np.append(self.y_tree, new_y)
        self.x_all = np.vstack((self.x_all, new_x))
        self.y_all = np.append(self.y_all, new_y)
        self.seed_all = np.append(self.seed_all, self.eval_seeds)
        min_ix = np.argmin(self.y_all)
        self.best_x = self.x_all[min_ix]
        self.best_y = self.y_all[min_ix]
        self.t_build_arr = np.append(self.t_build_arr, self.t_build) 
        self.t_srs_arr = np.append(self.t_srs_arr, self.t_srs)
        self.t_prop_arr = np.append(self.t_prop_arr, self.t_prop)
        try:
            self.t_eval_arr = np.append(self.t_eval_arr, self.t_eval)
        except:
            # i.e., self.t_eval is not defined. This could happen when one uses customized evaluation function.
            self.t_eval_arr = np.append(self.t_eval_arr, np.nan)
        self.gSRS_pct_arr = np.append(self.gSRS_pct_arr, self.gSRS_pct)
        self.zoom_lv_arr = np.append(self.zoom_lv_arr, self.zoom_lv)
           
        if self.i_iter_doe < self._n_iter_doe: # i.e., current iteration is in DOE phase
            self.i_iter_doe += 1
        else:
            # update weight pattern in SRS method
            if self._n_worker == 1:
                self.srs_wgt_pat = np.array([self._wgt_pat_bd[0]]) if self.srs_wgt_pat[0] == self._wgt_pat_bd[1] \
                    else np.array([self._wgt_pat_bd[1]]) # alternating weights
            # update tree node
            npt = len(self.x_tree)
            self.act_node['ix'] = np.append(self.act_node['ix'], np.arange(npt-self._n_worker, npt, dtype=int))
            if self._n_worker > 1 or (self._n_worker == 1 and self.srs_wgt_pat[0] == self._wgt_pat_bd[0]):
                if self.p_val >= 0.1:
                    # compute p_val
                    if self._use_eff_npt:
                        eff_n = eff_npt(self.x_tree[self.act_node['ix']], self.act_node['domain'])
                    else:
                        eff_n = len(self.x_tree[self.act_node['ix']])
                    self.p_val = self.p_val*eff_n**(-self._alpha/float(self._dim))
                    
                if self.gSRS_pct == 0: # i.e. pure local SRS
                    best_Y_prev = np.min(self.y_node)
                    best_Y_new = np.min(new_y) # minimum of Y values of newly proposed points
                    if best_Y_prev <= best_Y_new: # failure                
                        self.n_fail += 1 # count failure
                    else:
                        self.n_fail = 0
                    if self.n_fail == self._max_C_fail:
                        self.n_fail = 0
                        self.gamma -= self._delta_gamma
                        self.n_reduce_sigma += 1 # update counter
                self.act_node['state']['p'] = self.p_val
                self.act_node['state']['Cr'] = self.n_reduce_sigma
                self.act_node['state']['Cf'] = self.n_fail
                self.act_node['state']['gamma'] = self.gamma
            
            if self.n_reduce_sigma > self._max_n_reduce_sigma:
                # then we either restart or zoom-in (i.e., critical state is reached)
                Y_fit = self.rbf_mod(self.x_tree[self.act_node['ix']])
                min_ix = np.argmin(Y_fit)
                x_star = self.x_tree[self.act_node['ix']][min_ix]
                # suppose we're going to zoom in
                child_node_ix = self.get_child_node(x_star)
                if child_node_ix is None:
                    # then we create a new child (if zoom in)
                    domain_lb, domain_ub = zip(*self.act_node['domain'])
                    blen = np.array(domain_ub)-np.array(domain_lb) # bound length for each dimension
                    assert(np.min(blen)>0)
                    domain_lb = np.maximum(x_star-self._rho/2.*blen, domain_lb)
                    domain_ub = np.minimum(x_star+self._rho/2.*blen, domain_ub)
                    domain = list(zip(domain_lb, domain_ub)) # the list function is used to ensure compatibility of python3
                    child_node = {'ix': np.nonzero(boxify(self.x_tree, domain)[0])[0], 
                                  'domain': domain,
                                  'parent_ix': self.act_node_ix,
                                  'beta': self._init_beta,
                                  'state': self.init_node_state()}
                else:
                    # then we activate an existing child node (if zoom in)
                    child_node = self.tree[self.zoom_lv+1][child_node_ix]
                    child_node['ix'] = np.nonzero(boxify(self.x_tree, child_node['domain'])[0])[0]
                child_npt = len(child_node['ix'])
                domain_lb, domain_ub = zip(*child_node['domain'])
                blen = np.array(domain_ub)-np.array(domain_lb) # bound length for each dimension
                assert(np.min(blen)>0)
                
                if np.all(blen*child_npt**(-1./self._dim) < (self._prob.domain_ub-self._prob.domain_lb)*self._resol): # resolution condition
                    # then we restart
                    if verbose:
                        specific_msg += 'Restart for the next iteration!\n'
                    self.i_iter_doe = 0
                    self.doe_samp = self.doe()
                    self.i_cycle += 1
                    self.zoom_lv = 0
                    self.act_node_ix = 0
                    self.x_tree = np.zeros((0, self._dim))
                    self.y_tree = np.zeros(0)
                    self.tree = self.init_tree()
                else:
                    # then we zoom in
                    self.act_node['state'] = self.init_node_state() # reset the state of the current node
                    self.zoom_lv += 1
                    if child_node_ix is None:
                        # then we create a new child
                        if self.zoom_lv not in self.tree.keys():
                            self.act_node_ix = 0
                            self.tree[self.zoom_lv] = [child_node]    
                        else:
                            self.act_node_ix = len(self.tree[self.zoom_lv])
                            self.tree[self.zoom_lv].append(child_node)    
                        if verbose:
                            specific_msg += 'Zoom in (created a new child node)!\n'    
                    else:
                        # then activate existing child node
                        self.act_node_ix = child_node_ix
                        # reduce zoom-out probability
                        child_node['beta'] = max(self._min_beta, child_node['beta']/2.)
                        if verbose:
                            specific_msg += 'Zoom in (activated an existing child node)!\n'
            
            if self._n_worker > 1 or (self._n_worker == 1 and self.srs_wgt_pat[0] == self._wgt_pat_bd[0]):
                if np.random.uniform() < self.tree[self.zoom_lv][self.act_node_ix]['beta'] and self.zoom_lv > 0 and self.i_iter_doe >= self._n_iter_doe:
                    # then we zoom out
                    child_node = self.tree[self.zoom_lv][self.act_node_ix]
                    self.act_node_ix = child_node['parent_ix']
                    self.zoom_lv -= 1
                    assert(self.act_node_ix is not None)
                    # check that the node after zooming out will contain the current node
                    assert(domain_intersect(self.tree[self.zoom_lv][self.act_node_ix]['domain'], child_node['domain']) == child_node['domain'])
                    if verbose:
                        specific_msg += 'Zoom out!\n'
                        
        t2 = default_timer()
        t_update = t2-t1
        self.t_update_arr = np.append(self.t_update_arr, t_update)
        if verbose:
            sys.stdout.write(' Done (time took: %.2e sec).\n' % t_update+specific_msg)
            
        self.save_state()
   
     
    def is_done(self):
        """
        Check whether we are done with the optimization.
        
        Returns:
            
            done (bool): Indicator.
        """
        if self._n_iter is None:
            assert(self.i_cycle <= self._n_cycle)
            done = self.i_cycle == self._n_cycle
        else:
            assert(self.i_iter <= self._n_iter)
            done = self.i_iter == self._n_iter
            
        return done
    
    
    def get_child_node(self, x_star):
        """
        Get the child node for the optimization tree.
        
        Args:
            
            x_star (1d array): Focal point of zoom-in.
            
        Returns:
            
            child_ix (int or None): Selected child node index. 
                If None, then we need to create a new child.
        """
        assert(x_star.ndim == 1)
        # get zoom level of a child node
        n_zoom_child = self.zoom_lv+1
        if n_zoom_child not in self.tree.keys():
            child_ix = None
        else:
            # the indice of candidate child nodes
            child_node_ix_list = [i for i, c in enumerate(self.tree[n_zoom_child]) if all(boxify(x_star.reshape((1, -1)), c['domain'])[0])]
            if len(child_node_ix_list) == 0:
                child_ix = None
            else:
                # find the child node among the candidates, of which the center of the domain is closest to x_star
                dist_list = [np.linalg.norm(np.mean(self.tree[n_zoom_child][i]['domain'])-x_star) for i in child_node_ix_list]
                child_ix = child_node_ix_list[np.argmin(dist_list)]
        
        return child_ix
    
    
    def save_state(self):
        """
        Save the state of the optimizer to files.
        """
        # save state to pickle file
        with open(self._state_pkl_temp_file, 'wb') as f:
            pickle.dump(np.random.get_state(), f) # first save to temporary file, preventing data loss due to termination during execution of `pickl.dump`
        shutil.copy2(self._state_pkl_temp_file, self._state_pkl_file) # create a new or overwrite the old `self._state_pkl_file`
        os.remove(self._state_pkl_temp_file) # remove temporary file
        
        # save state to npz file
        np.savez(self._state_npz_temp_file,
                 # constant parameters
                 _dim=self._dim, _n_worker=self._n_worker, _n_iter=self._n_iter, _n_cycle = self._n_cycle,
                 _resume=self._resume, _seed=self._seed, _parallel_training=self._parallel_training, _out_dir=self._out_dir, 
                 _n_cand_fact=self._n_cand_fact, _wgt_pat_bd=self._wgt_pat_bd, _normalize_data=self._normalize_data,
                 _init_gamma=self._init_gamma, _delta_gamma=self._delta_gamma, _init_sigma=self._init_sigma, 
                 _max_n_reduce_sigma=self._max_n_reduce_sigma, _rho=self._rho, _init_p=self._init_p,
                 _init_beta=self._init_beta, _min_beta=self._min_beta, _alpha=self._alpha, _lambda_range=self._lambda_range,
                 _rbf_kernel=self._rbf_kernel, _rbf_poly_deg=self._rbf_poly_deg, _n_fold=self._n_fold, _resol=self._resol,
                 _use_eff_npt=self._use_eff_npt, _max_C_fail=self._max_C_fail, _n_iter_doe=self._n_iter_doe,
                 _n_doe_samp=self._n_doe_samp, _n_cand=self._n_cand, _state_npz_file=self._state_npz_file,
                 _state_pkl_file=self._state_pkl_file, _state_npz_lock_file=self._state_npz_lock_file,
                 _state_pkl_lock_file=self._state_pkl_lock_file, _state_npz_temp_file=self._state_npz_temp_file,
                 _state_pkl_temp_file=self._state_pkl_temp_file, _verbose_dot_len=self._verbose_dot_len,
                 # state variables
                 i_iter=self.i_iter, i_cycle=self.i_cycle, doe_samp=self.doe_samp, i_iter_doe=self.i_iter_doe,
                 t_build_arr=self.t_build_arr, t_srs_arr=self.t_srs_arr, t_prop_arr=self.t_prop_arr,
                 t_eval_arr=self.t_eval_arr, t_update_arr=self.t_update_arr, gSRS_pct_arr=self.gSRS_pct_arr,
                 zoom_lv_arr=self.zoom_lv_arr, x_tree=self.x_tree, y_tree=self.y_tree, x_all=self.x_all,
                 y_all=self.y_all, seed_all=self.seed_all, best_x=self.best_x, best_y=self.best_y, 
                 zoom_lv=self.zoom_lv, act_node_ix=self.act_node_ix, srs_wgt_pat=self.srs_wgt_pat, tree=self.tree,
                 eval_seeds=self.eval_seeds)
        
        shutil.copy2(self._state_npz_temp_file, self._state_npz_file)
        os.remove(self._state_npz_temp_file) # remove temporary file
   
    
    def load_state(self):
        """
        Load the state of the optimizer from files.
        """        
        # load state data from pkl file
        if os.path.isfile(self._state_pkl_lock_file):
            os.remove(self._state_pkl_lock_file) # remove lock file, if there's any
        with open(self._state_pkl_file, 'rb') as f:
            np.random.set_state(pickle.load(f))    
        # load state data from npz file
        if os.path.isfile(self._state_npz_lock_file):
            os.remove(self._state_npz_lock_file) # remove lock file, if there's any
        data = np.load(self._state_npz_file)
        # check consistency
        assert(self._dim==data['_dim'] and self._n_worker==data['_n_worker'] and self._n_cand_fact==data['_n_cand_fact']
               and np.all(self._wgt_pat_bd==data['_wgt_pat_bd']) and self._normalize_data==data['_normalize_data']
               and self._init_gamma==data['_init_gamma'] and self._delta_gamma==data['_delta_gamma']
               and self._init_sigma==data['_init_sigma'] and self._max_n_reduce_sigma==data['_max_n_reduce_sigma']
               and self._rho==data['_rho'] and self._init_p==data['_init_p'] and self._init_beta==data['_init_beta']
               and self._min_beta==data['_min_beta'] and self._alpha==data['_alpha']
               and np.all(self._lambda_range==data['_lambda_range']) and np.all(self._rbf_kernel==data['_rbf_kernel'])
               and self._rbf_poly_deg==data['_rbf_poly_deg'] and self._n_fold==data['_n_fold'] and self._resol==data['_resol']
               and self._use_eff_npt==data['_use_eff_npt'] and self._max_C_fail==data['_max_C_fail']
               and self._n_cand==data['_n_cand']), 'To resume, experiment condition needs to be consistent with that of the last run.'
        # read state variables
        self.i_iter = data['i_iter'].item(0)
        self.i_cycle = data['i_cycle'].item(0)
        self.doe_samp = data['doe_samp']
        self.i_iter_doe = data['i_iter_doe'].item(0)
        self.t_build_arr = data['t_build_arr']
        self.t_srs_arr = data['t_srs_arr']
        self.t_prop_arr = data['t_prop_arr']
        self.t_eval_arr = data['t_eval_arr']
        self.t_update_arr = data['t_update_arr']
        self.gSRS_pct_arr = data['gSRS_pct_arr']
        self.zoom_lv_arr = data['zoom_lv_arr']
        self.x_tree = data['x_tree']
        self.y_tree = data['y_tree']
        self.x_all = data['x_all']
        self.y_all = data['y_all']
        self.seed_all = data['seed_all']
        self.best_x = data['best_x']
        self.best_y = data['best_y'].item(0)
        self.zoom_lv = data['zoom_lv'].item(0)
        self.act_node_ix = data['act_node_ix'].item(0)
        self.srs_wgt_pat = data['srs_wgt_pat']
        self.tree = data['tree'].item(0)
        self.eval_seeds = data['eval_seeds']
        # sanity check
        if self._n_iter is None:
            assert(self._n_cycle > self.i_cycle), 'In the last run, %d optimization cycles were completed. To resume, please set n_cycle greater than %d (n_cycle is %d currently).' \
                % (self.i_cycle, self.i_cycle, self._n_cycle)
        else:
            assert(self._n_iter > self.i_iter), 'In the last run, %d iterations were completed. To resume, please set n_iter greater than %d (n_iter is %d currently).' \
                % (self.i_iter, self.i_iter, self._n_iter)
            
        
    def visualize(self, select=['optim_curve', 'zoom_level', 'time'], save_file=None):
        """
        Visualize optimization progress.
        
        Args:
            
            select (list or tuple, optional): Select which plot(s) to show.
                Possible values and their meanings:
                    'optim_curve': optimization curve (value of best point vs. iteration)
                    'zoom_level': Zoom level plot (zoom level vs. iteration)
                    'time': Wall time of different components (time vs. iteration)
                    
            save_file (list or tuple or None, optional): File paths to save figures.
                If ``list`` or ``tuple``, then it consists of file paths for each plot in `select`.
                If None, then no figures will be saved.
        """
        select_possible_vals = ['optim_curve', 'zoom_level', 'time']
        assert(type(select) in [list, tuple] and set(select)<=set(select_possible_vals)), 'invalid select value'
        if save_file is not None:
            assert(type(save_file) in [list, tuple] and len(save_file) == len(select)), 'invalid save_file value'
            file_dict = {x:y for x, y in zip(select, save_file)}
        else:
            file_dict = {}
        iterations = range(1, self.i_iter+1)
        
        try:
            if 'optim_curve' in select:
                vals = self.y_all.reshape((self.i_iter, self._n_worker))
                best_val_arr = np.zeros(self.i_iter)
                for i, val in enumerate(vals):
                    if i == 0:
                        best_val_arr[i] = np.amin(val)
                    else:
                        best_val_arr[i] = min(best_val_arr[i-1], np.amin(val))
                fig = plt.figure()
                plt.plot(iterations, best_val_arr, 'b-', label='best (noisy) value')
                if self._prob.min_true_func is not None:
                    plt.plot(iterations, np.ones_like(iterations)*self._prob.min_true_func, '-', label='global min')
                    if self._prob.sd is not None:
                        plt.plot(iterations, np.ones_like(iterations)*(self._prob.min_true_func+self._prob.sd), 
                                 '--', label='1 std of noise')
                    plt.legend(loc='best', framealpha=0.5)
                plt.grid(True)
                plt.xlabel('Iteration')
                plt.ylabel('Function value')
                plt.title('Optimization curve (problem: %s)' % self._prob.name)
                fig.gca().xaxis.set_major_locator(MaxNLocator(integer=True)) # force x ticks to be integer-valued
                plt.show()
                if 'optim_curve' in file_dict: 
                    fig.savefig(file_dict['optim_curve'])
                    
            if 'zoom_level' in select:
                fig = plt.figure()
                plt.plot(iterations, self.zoom_lv_arr, 'b-')
                plt.grid(True)
                plt.xlabel('Iteration')
                plt.ylabel('Zoom level')
                plt.ylim([-1, np.amax(self.zoom_lv_arr)+1])
                plt.yticks(np.arange(np.amax(self.zoom_lv_arr)+1))
                plt.title('Zoom level (problem: %s)' % self._prob.name)
                fig.gca().xaxis.set_major_locator(MaxNLocator(integer=True)) # force x ticks to be integer-valued
                plt.show()
                if 'zoom_level' in file_dict:
                    fig.savefig(file_dict['zoom_level'])
            
            if 'time' in select:
                fig = plt.figure()
                plt.plot(iterations, self.t_build_arr, '-', label='build RBF')
                plt.plot(iterations, self.t_srs_arr, '-', label='SRS')
                if not np.all(np.isnan(self.t_eval_arr)):
                    plt.plot(iterations, self.t_eval_arr, '-', label='evaluation')
                plt.plot(iterations, self.t_update_arr, '-', label='update')
                plt.grid(True)
                plt.xlabel('Iteration')
                plt.ylabel('Wall time (sec)')
                plt.yscale('log')
                plt.legend(loc='best', framealpha=0.5)
                plt.title('Computational cost (problem: %s)' % self._prob.name)
                fig.gca().xaxis.set_major_locator(MaxNLocator(integer=True)) # force x ticks to be integer-valued
                plt.show()
                if 'time' in file_dict:
                    fig.savefig(file_dict['time'])
        except Exception as e:
            sys.exit('Error! Unable to generate plots for visualization: %s. This may be due to unsuccessful installation of matplotlib package. For more, please see the installation note in the README file at `https://github.com/compdyn/ProSRS`.' % str(e))
    
    
    def posterior_eval(self, n_top=0.1, n_repeat=10, n_worker=None, seed=1, seed_func=None, verbose=True):
        """
        Posterior Monte Carlo evaluations for selecting the true best point.
        
        Args:
            
            n_top (float or int, optional): Proportion/number of top points to be evaluated.
                If ``float``, then `n_top` in (0, 1) is the proportion of top points
                among all the evaluated points. If ``int``, then `n_top`, a positive
                integer, is the number of top points to be evaluated.
            
            n_repeat (int, optional): Number of Monte Carlo repeats for the evaluations.
            
            n_worker (int or None, optional): Number of workers for the evaluations.
                Used for parallel computing. If None, then we use the default value = 
                `self._n_worker`. Here we assume parallelism across multiple cores
                in a machine and using Python package ``pathos``.
            
            seed (int, optional): (Base) random seed for Monte Carlo evaluations.
                For a point with point index i (zero-based) and repeat index j 
                (zero-based), the seed for the point = `seed`+i*`n_repeat`+j.
            
            seed_func (callable or None, optional): User-specified function for setting the random seed.
                If None, then we use ``numpy.random.seed`` method to set random seed.
                If callable, it is a function taking ``seed`` as an argument.
                Note: using ``numpy.random.seed`` may not always gaurantee
                the reproducibility. Here we provide an option for users to specify their own routines.
            
            verbose (bool, optional): Whether to verbose about posterior evaluations.
        """        
        # sanity check
        assert(type(n_top) in [float, int])
        assert(type(n_repeat) is int and n_repeat > 0)
        assert(type(seed) is int and seed >= 0)
        n_all = len(self.x_all)
        assert(n_all > 0), 'No points have been evaluated yet. Run ProSRS algorithm first before running posterior evaluations.'
        if type(n_top) is float:
            assert(0 < n_top < 1)
            n_top = max(1, int(round(n_top*n_all)))
        else:
            assert(n_top > 0)
            if n_top > n_all:
                warnings.warn('Specified n_top is larger than total number of points (%d). Reset n_top = %d.' % (n_all, n_all))
                n_top = n_all
        n_worker = self._n_worker if n_worker is None else n_worker
        assert(type(n_worker) is int and n_worker > 0)
        
        t1 = default_timer()
        if verbose:
            sys.stdout.write('Running posterior evaluations (n_point = %d, n_repeat = %d) ' % (n_top, n_repeat)+'.'*self._verbose_dot_len)
            
        # evaluate top points
        sort_ix = np.argsort(self.y_all)  
        top_pt = self.x_all[sort_ix][:n_top]
        top_pt_mc = top_pt[np.outer(range(n_top), np.ones(n_repeat, dtype=int)).flatten()] # duplicate each point in `top_pt` for `n_repeat` times
        top_seed_mc = seed+np.arange(n_top*n_repeat, dtype=int)
        assert(len(top_pt_mc) == len(top_seed_mc))
        top_val_mc = eval_func(self._prob.f, top_pt_mc, n_proc=n_worker, seeds=top_seed_mc.tolist(), seed_func=seed_func)
        top_val = top_val_mc.reshape((n_top, n_repeat))
        # save results
        self.i_iter_posterior_eval = self.i_iter
        self.posterior_eval_x = top_pt
        self.posterior_eval_y = top_val
        self.posterior_mean = np.mean(self.posterior_eval_y, axis=1)
        self.posterior_std = np.std(self.posterior_eval_y, axis=1, ddof=1)
        min_ix = np.argmin(self.posterior_mean)
        self.true_best_x = self.posterior_eval_x[min_ix]
        self.true_best_y = self.posterior_mean[min_ix]
        
        t2 = default_timer()
        if verbose:
            sys.stdout.write(' Done (time took: %.2e sec).\n' % (t2-t1))
        



"""
Copyright (C) 2016-2019 Chenchao Shou

Licensed under Illinois Open Source License (see the file LICENSE). For more information
about the license, see http://otm.illinois.edu/disclose-protect/illinois-open-source-license.

Contain functions and class that deal with the surrogate.
"""
import numpy as np
from scipy.special import xlogy
from scipy import linalg
import warnings
import multiprocessing as mp
from sklearn import preprocessing
from sklearn.model_selection import KFold
from functools import partial
from ..utility.functions import unique_row


class Rbf(object):
    """
    A class dealing with L2-regularized RBF regression.
    
    The code is inherited from `scipy.interpolate.Rbf <https://github.com/scipy/scipy/blob/v0.19.0/scipy/interpolate/rbf.py>`,
    and is further developed by Chenchao Shou.
    
    rbf - Radial basis functions for interpolation/smoothing scattered Nd data.
    Written by John Travers <jtravs@gmail.com>, February 2007
    Based closely on Matlab code by Alex Chirokov
    Additional, large, improvements by Robert Hetland
    Some additional alterations by Travis Oliphant
    Permission to use, modify, and distribute this software is given under the
    terms of the SciPy (BSD style) license.  See LICENSE.txt that came with
    this distribution for specifics.
    NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
    Copyright (c) 2006-2007, Robert Hetland <hetland@tamu.edu>
    Copyright (c) 2007, John Travers <jtravs@gmail.com>

    A class for radial basis function approximation/interpolation of
    n-dimensional scattered data using L2 regularization.
    Parameters
    ----------
    *args : arrays
        x, y, z, ..., d, where x, y, z, ... are the coordinates of the nodes
        and d is the array of values at the nodes
    function : str, optional
        The radial basis function, based on the radius, r, given by the norm
        (default is Euclidean distance); the default is 'multiquadric'::
            'multiquadric': sqrt((r/self.epsilon)**2 + 1)
            'inverse': 1.0/sqrt((r/self.epsilon)**2 + 1)
            'gaussian': exp(-(r/self.epsilon)**2)
            'linear': r
            'cubic': r**3
            'quintic': r**5
            'thin_plate': r**2 * log(r)
    epsilon : float, optional
        Adjustable constant for gaussian or multiquadrics functions
        - defaults to approximate average distance between nodes (which is
        a good start).
    smooth : float, optional
        Values greater than zero increase the smoothness of the
        approximation.  0 is for interpolation (default), the function will
        always go through the nodal points in this case.
    norm : callable, optional
        A function that returns the 'distance' between two points, with
        inputs as arrays of positions (x, y, z, ...), and an output as an
        array of distance.  E.g, the default::
            def euclidean_norm(x1, x2):
                return sqrt( ((x1 - x2)**2).sum(axis=0) )
        which is called with x1=x1[ndims,newaxis,:] and
        x2=x2[ndims,:,newaxis] such that the result is a matrix of the
        distances from each point in x1 to each point in x2.
    use_scipy_rbf: boolean, optional
        Whether to use scipy rbf. Default: False.
    suppress_warning: boolean, optional
        Whether to suppress warnings when doing linalg.solve. Default: True.
    
    """

    def _euclidean_norm(self, x1, x2):
        return np.sqrt(((x1 - x2)**2).sum(axis=0))

    def _h_multiquadric(self, r):
        return np.sqrt((1.0/self.epsilon*r)**2 + 1)

    def _h_inverse_multiquadric(self, r):
        return 1.0/np.sqrt((1.0/self.epsilon*r)**2 + 1)

    def _h_gaussian(self, r):
        return np.exp(-(1.0/self.epsilon*r)**2)

    def _h_linear(self, r):
        return r

    def _h_cubic(self, r):
        return r**3

    def _h_quintic(self, r):
        return r**5

    def _h_thin_plate(self, r):
        return xlogy(r**2, r)

    # Setup self._function and do smoke test on initial r
    def _init_function(self, r):
        if isinstance(self.function, str):
            self.function = self.function.lower()
            _mapped = {'inverse': 'inverse_multiquadric',
                       'inverse multiquadric': 'inverse_multiquadric',
                       'thin-plate': 'thin_plate'}
            if self.function in _mapped:
                self.function = _mapped[self.function]

            func_name = "_h_" + self.function
            if hasattr(self, func_name):
                self._function = getattr(self, func_name)
            else:
                functionlist = [x[3:] for x in dir(self) if x.startswith('_h_')]
                raise ValueError("function must be a callable or one of " +
                                     ", ".join(functionlist))
            self._function = getattr(self, "_h_"+self.function)

        a0 = self._function(r)
        if a0.shape != r.shape:
            raise ValueError("Callable must take array and return array of the same shape")
        return a0

    def __init__(self, *args, **kwargs):
        self.xi = np.asarray([np.asarray(a, dtype=np.float_).flatten()
                           for a in args[:-1]])
        self.N = self.xi.shape[-1]
        self.di = np.asarray(args[-1]).flatten()

        if not all([x.size == self.di.size for x in self.xi]):
            raise ValueError("All arrays must be equal length.")

        self.norm = kwargs.pop('norm', self._euclidean_norm)
        r = self._call_norm(self.xi, self.xi)
        self.epsilon = kwargs.pop('epsilon', None)
        if self.epsilon is None:
            ximax = np.amax(self.xi, axis=1)
            ximin = np.amin(self.xi, axis=1)
            edges = ximax-ximin
            edges = edges[np.nonzero(edges)]
            self.epsilon = np.power(np.prod(edges)/self.N, 1.0/edges.size)
        self.smooth = kwargs.pop('smooth', 0.0)
        self.function = kwargs.pop('function', 'multiquadric')        
        
        assert(self.smooth > 0),'smoothing parameter needs to be positive'
        self.use_scipy_rbf = kwargs.pop('use_scipy_rbf', False)
        self.suppress_warning = kwargs.pop('suppress_warning', True)
        self.wgt = kwargs.pop('wgt',np.ones(self.N)) # weight vector        
        assert ((len(self.wgt) == self.N) and np.min(self.wgt)>=0), 'invalid weight vector'
        W = np.diag(self.wgt) # construct weight matrix
        self.deg = kwargs.pop('deg',0) # degree of polynomials (if = 0, then no polynomial tail)
        assert(self.deg in [0,1]), 'currently we only support degree up to 1'
        
        with warnings.catch_warnings():
            if self.suppress_warning:
                warnings.simplefilter('ignore')
            else:
                warnings.simplefilter('default')
            # attach anything left in kwargs to self
            #  for use by any user-callable function or
            #  to save on the object returned.
            for item, value in kwargs.items():
                setattr(self, item, value)
            if self.use_scipy_rbf:
                self.A = self._init_function(r) - np.eye(self.N)*self.smooth
                self.nodes = linalg.solve(self.A, self.di)
            else:
                phi = self._init_function(r)
                assert(phi.shape == (self.N, self.N))
                if self.deg == 1:
                   phi = np.hstack((phi,self.xi.T))
                   phi = np.hstack((phi,np.ones((self.N,1))))
                self.A = np.dot(np.dot(phi.T,W),phi)+np.eye(phi.shape[1])*self.smooth
                self.b = np.dot(np.dot(phi.T,W),self.di)
                # since smoothing parameter is positive, A must be positive definite.
                # so we can set sym_pos=True when solving with linalg.solve
                self.nodes = linalg.solve(self.A,self.b,sym_pos=True,lower=True)

    def _call_norm(self, x1, x2):
        if len(x1.shape) == 1:
            x1 = x1[np.newaxis, :]
        if len(x2.shape) == 1:
            x2 = x2[np.newaxis, :]
        x1 = x1[..., :, np.newaxis]
        x2 = x2[..., np.newaxis, :]
        return self.norm(x1, x2)

    def __call__(self, *args):
        args = [np.asarray(x) for x in args]
        if not all([x.shape == y.shape for x in args for y in args]):
            raise ValueError("Array lengths must be equal")
        shp = args[0].shape
        xa = np.asarray([a.flatten() for a in args], dtype=np.float_)        
        r = self._call_norm(xa, self.xi) 
        if self.deg == 0:
            y_val = np.dot(self._function(r), self.nodes).reshape(shp)
        elif self.deg == 1:
            nx = xa.shape[1] # number of points to be evaluated
            phi = self._function(r)
            assert(phi.shape == (nx, self.N) and self.xi.shape[0] == xa.shape[0])
            phi = np.hstack((phi,xa.T))
            phi = np.hstack((phi,np.ones((nx,1))))
            y_val = np.dot(phi, self.nodes).reshape(shp)
        return y_val
    
    
def RBF_reg(X, Y, sm_range, normalize_data=True, wgt_expon=0., use_scipy_rbf=False,
            log_opt=True, n_fold=5, n_min_sm=10, kernel='multiquadric', poly_deg=0,
            pool=None):
    """
    Construct a RBF regression model using L2 regularization.
    
    Args:
        
        X (2d array): Data points (each row is one point).
        
        Y (1d array): (Noisy) function evaluation of each point in `X`.
        
        sm_range (list or tuple or float): Range of smoothing parameter (L2 regularization parameter).
            If ``list`` or ``tuple``, then `sm_range` = [lower bound, upper bound].
            The value of the smoothing parameter will be determined via a cross validation procedure
            on the domain `sm_range`. If ``float``, then the smoothing parameter is equal to
            `sm_range` (no cross validation in this case).
            
        normalize_data (bool, optional): Whether to normalize data before training a RBF model.
        
        wgt_expon (float, optional): Weight exponent in RBF regression ('gamma').
            Useful only when `use_scipy_rbf` = False.
        
        use_scipy_rbf (bool, optional): Whether to use Scipy implementation for the RBF model.
        
        log_opt (bool, optional): Whether to optimize the smoothing parameter on a log scale 
            during cross validation.
        
        
        n_fold (int, optional): Number of folds for the cross validation.
        
        n_min_sm (int, optional): Minimum number of samples on the domain `sm_range`.
            The value of the smoothing parameter will be selected to be the one, among
            these samples, of which the cross-validated error is the lowest.
        
        kernel (str, optional): RBF kernel. For the possible values and their meanings, see
            `scipy.interpolate.Rbf <https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.interpolate.Rbf.html>`.
            
        poly_deg (int, optional): Degree of RBF polynomial tail (either 0 or 1). 
            If zero, then no polynomial tail. Useful only when `use_scipy_rbf` = False.
        
        pool (`pathos.multiprocessing.ProcessingPool` type or None): Parallel pool object.
            If None, then we perform training in serial. Otherwise, we use `pathos.multiprocessing`
            module for parallelism. Our tests have shown that depending on the machine and the 
            optimization function, sometimes parallelism may cause memory issues. So we set ``pool=None`` by default.
    
    Returns:
        
        rbf_mod (callable): RBF regression model.
        
        opt_sm (float): (Optimal) smoothing parameter found through cross validation.
        
        cv_err (list or None): cross validated error for each smoothing parameter sample.
            If None, then there is no cross validation.
        
        rbf_obj (`Rbf` type or None): Original RBF instance before normalization. 
            If None, then the model is degenerated to a constant function.
    """
    # sanity check
    assert(len(X) == len(Y) > 0)
    assert(len(sm_range) in [1, 2])
    assert(type(n_fold) is int and n_fold > 1)
    
    n_uniq = len(unique_row(X)) # find number of unique X
    if n_uniq > 1:
        # then we build RBF surrogate
        if normalize_data:
            # normalize X and Y to [0,1]
            X_scaler = preprocessing.MinMaxScaler()
            Y_scaler = preprocessing.MinMaxScaler()
            X = X_scaler.fit_transform(X)
            Y = Y_scaler.fit_transform(Y.reshape((-1, 1))).flatten()
        else:
            X_scaler = Y_scaler = None 
            
        if len(sm_range) == 2:
            sm_lw, sm_up = sm_range
            assert(sm_lw < sm_up), 'invalid sm_range'
            
            n_proc = mp.cpu_count() # number of available cores
            n_sm = n_proc*int(np.ceil(n_min_sm/float(n_proc))) # number of candidate smoothing parameters (multiple of `n_proc` so efficiency of parallelism is maximized)
            assert(n_sm >= n_min_sm > 0) # sanity check
            # find candidate smoothing parameters
            if log_opt:
                smooth = np.logspace(np.log10(sm_lw), np.log10(sm_up), n_sm)
            else:
                smooth = np.logspace(sm_lw, sm_up, n_sm)
            if pool is not None:
                assert(pool.nodes == n_proc)
                CV_smooth_partial = partial(CV_smooth, X=X, Y=Y, n_fold=n_fold, kernel=kernel,
                                            use_scipy_rbf=use_scipy_rbf, wgt_expon=wgt_expon, poly_deg=poly_deg)
                cv_err = pool.map(CV_smooth_partial, smooth)
            else:
                cv_err = [CV_smooth(s, X, Y, n_fold, kernel, use_scipy_rbf, wgt_expon, poly_deg) for s in smooth]
            # find smoothing parameter that has the lowest cross-validated error
            opt_ix = np.argmin(cv_err)
            opt_sm = smooth[opt_ix]
            
        else: # i.e., we specify the opt_sm
            opt_sm = sm_range[0]
            cv_err = None
            
        # then we build RBF surrogate with the optimal smoothing parameter using all the data
        XY_data = np.vstack((X.T, Y.reshape((1, -1))))
        if not use_scipy_rbf:
            wgt = rbf_wgt(Y, wgt_expon)
            rbf_obj = Rbf(*XY_data, wgt=wgt, function=kernel, smooth=opt_sm, deg=poly_deg)        
        else:
            rbf_obj = Rbf(*XY_data, use_scipy_rbf=use_scipy_rbf, function=kernel, smooth=opt_sm)
        rbf_mod = normalize_rbf(rbf_obj, X_scaler, Y_scaler)
        
    else:
        # pathological case: X overlaps. Then the model is simply the mean of y values.
        opt_sm, cv_err, rbf_obj = np.nan, None, None
        def rbf_mod(x):
            """
            Degenerated RBF model.
            
            Args:
                
                x (2d array): Points to be evaluated. Each row is one point.
            
            Returns:
                
                y (1d array): Function evaluations at points in `x`.
            """
            assert(x.ndim == 2)
            nx = len(x)
            y = np.mean(Y)*np.ones(nx)
            
            return y
        
    return rbf_mod, opt_sm, cv_err, rbf_obj


def normalize_rbf(rbf_obj, X_scaler, Y_scaler):
    """
    Generate a normalized RBF model.
    
    Args:
        
        rbf_obj (`Rbf` type): Original RBF model.
        
        X_scaler (`sklearn.preprocessing` scaler type or None): Scaler for input X. 
            If None, then no normalization will be performed on the model input.
                    
        Y_scaler (`sklearn.preprocessing` scaler type or None): Scaler for output Y. 
            If None, then no normalization will be performed on the model output.
            
    Returns:
        
        norm_rbf_mod (callable): Normalized Rbf model.
    """
    def norm_rbf_mod(X):
        """
        Normalized RBF model.
        
        Args:
            
            X (2d array): Points to be evaluated. Each row is one point.
        
        Returns:
            
            Y (1d array): Function evaluations at points in `X`.
        """
        assert(X.ndim == 2)
        
        if X_scaler is not None:
            X = X_scaler.transform(X)
        Y = rbf_obj(*X.T)
        if Y_scaler is not None:
            Y = Y_scaler.inverse_transform(Y.reshape((-1, 1))).flatten()
            
        return Y
    
    return norm_rbf_mod


def rbf_wgt(Y, wgt_expon):
    """
    Generate weights of training data for RBF regression.
    
    Args:
        
        Y (1d array): Y values of training data.
        
        wgt_expon (float): Weight exponent ('gamma').
    
    Returns:
        
        wgt (1d array): Weights of training data.
    """
    assert(Y.ndim == 1)
    
    min_Y, max_Y = np.min(Y), np.max(Y)
    if min_Y == max_Y:
        Y = np.zeros_like(Y)
    else:
        Y = (Y-min_Y)/(max_Y-min_Y) # normalization
    assert (np.min(Y) >= 0 and np.max(Y) <= 1) # sanity check
    wgt = np.exp(wgt_expon*Y)
    
    return wgt
 
       
def CV_smooth(smooth, X, Y, n_fold, kernel, use_scipy_rbf, wgt_expon, poly_deg):
     """
     Get cross validated error for a smoothing parameter.
     
     Args:
         
         smooth (float): Smoothing parameter (L2 regularization constant).
         
         X (2d array): X values of training data. Each row is one point.
         
         Y (1d array): Y values of training data (1d array).
         
         n_fold (int): Number of folds for cross validation.
         
         kernel (str): RBF kernel. For the possible values and their meanings, see
            `scipy.interpolate.Rbf <https://docs.scipy.org/doc/scipy-0.19.0/reference/generated/scipy.interpolate.Rbf.html>`.
         
         use_scipy_rbf (bool): Whether to use Scipy implementation for the RBF model.
         
         wgt_expon (float): Weight exponent in RBF regression ('gamma').
            Useful only when `use_scipy_rbf` = False.
     
         poly_deg (int): Degree of RBF polynomial tail (either 0 or 1). 
            If zero, then no polynomial tail. Useful only when `use_scipy_rbf` = False.
     
     Returns:
         
         cv_err (float): Cross validated error.
     """
     nX = len(X)
     n_fold = min(nX, n_fold)
     assert(n_fold > 1) # sanity check
     
     kf = KFold(n_splits=n_fold)
     test_err = np.zeros(n_fold)
     for f, index in enumerate([j for j in kf.split(np.ones(nX))]):
        train_ix, test_ix = index
        X_train, X_test = X[train_ix], X[test_ix]
        Y_train, Y_test = Y[train_ix], Y[test_ix]
        # get number of unique X_train
        n_uniq = len(unique_row(X_train))
        if n_uniq > 1:
            # then we build RBF model
            XY_data = np.vstack((X_train.T, Y_train.reshape((1, -1))))
            if not use_scipy_rbf:
                wgt = rbf_wgt(Y_train, wgt_expon)
                rbf_obj = Rbf(*XY_data, wgt=wgt, function=kernel, smooth=smooth, deg=poly_deg)
            else:
                rbf_obj = Rbf(*XY_data, use_scipy_rbf=use_scipy_rbf, function=kernel, smooth=smooth)
            test_err[f], _ = eval_rbf_mod(X_test, Y_test, rbf_obj)
        else:
            # pathological case: X_train overlaps. Then the model is simply the mean of y values.
            Y_pred = np.mean(Y_train)
            error = Y_pred - Y_test
            test_err[f] = np.sqrt(np.mean(error**2)) # RMSE
     cv_err = np.mean(test_err)
     
     return cv_err


def eval_rbf_mod(X, Y, rbf_obj):
    """
    Evaluate performance of a RBF model on a test dataset.
    
    Args:
        
        X (2d array): X values of test data. Each row is one point.
         
        Y (1d array): Y values of test data (1d array).
        
        rbf_obj (`Rbf` type): RBF model to be evaluated.
    
    Returns:
        
        RMSE (float): Root mean square error on the test dataset.
        
        error (1d array): Prediction errors on the test dataset.
    """
    assert(X.ndim == 2 and Y.ndim == 1)
    
    error = rbf_obj(*X.T)-Y
    RMSE = np.sqrt(np.mean(error**2))
    
    return RMSE, error